
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__          import annotations

import base58
import codecs
import hashlib
import itertools
import json
import logging
import math
import re
import secrets
import string
import warnings

from functools		import wraps
from collections	import namedtuple
from typing		import Dict, List, Sequence, Tuple, Optional, Union, Callable

from shamir_mnemonic	import generate_mnemonics
from shamir_mnemonic.shamir import RANDOM_BYTES

import hdwallet
from hdwallet		import cryptocurrencies

from .defaults		import (
    BITS_DEFAULT, BITS, MNEM_ROWS_COLS, GROUPS, GROUP_REQUIRED_RATIO, GROUP_THRESHOLD_RATIO, CRYPTO_PATHS
)
from .util		import ordinal, commas, is_listlike, is_mapping
from .recovery		import produce_bip39, recover_bip39, recover as recover_slip39

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( __package__ )


# Support for private key encryption via BIP-38 and Ethereum JSON wallet is optional; pip install slip39[wallet]
paper_wallet_issues		= []
try:
    from Crypto.Cipher	import AES
    from Crypto.Protocol.KDF import scrypt
except ImportError as exc:
    AES				= None
    scrypt			= None
    message			= f"Unable to support Paper Wallet output: {exc}"
    warnings.warn( message, ImportWarning )
    if log.isEnabledFor( logging.DEBUG ):
        log.exception( message )
    paper_wallet_issues.append( message )

try:
    import eth_account
except ImportError as exc:
    eth_account			= None
    message			= f"Unable to support Paper Wallet output: {exc}"
    warnings.warn( message, ImportWarning )
    if log.isEnabledFor( logging.DEBUG ):
        log.exception( message )
    paper_wallet_issues.append( message )


def paper_wallet_available():
    """Determine if encrypted BIP-38 and Ethereum JSON Paper Wallets are available."""
    available			= AES and scrypt and eth_account
    if not available:
        log.warning( f"Paper Wallets unavailable; perhaps run: 'python3 -m pip install slip39[gui,wallet]': {', '.join( paper_wallet_issues )}" )
    return available


def path_edit(
    path: str,
    edit: str,
):
    """Replace the current path w/ the new path, either entirely, or if only partially if a
    continuation of dot(s) followed by some new path segment(s) is provided.  For example, if the
    default path for your desired format is:

        m/49'/0'/0'/0/0

    you can provide "../1/9" to produce "m/49'/0'/0'/1/9", or "...3" to produce "m/49'/0'/0'/0/3".

    To simply eliminate a segment, leave its segment empty.  For example, if you wish to eliminate
    the last two segments, provide "..//", yielding "m/49'/0'/0'".

    """
    if edit.startswith( '.' ):
        if ( new_edit := edit.lstrip( '.' )).startswith( '/' ):
            new_edit	= new_edit[1:]
        new_segs	= new_edit.split( '/' )
        cur_segs	= path.split( '/' )
        log.debug( f"Using {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        if len( new_segs ) >= len( cur_segs ):
            raise ValueError( f"Cannot use {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        # Truncate the number of edited segs, and appends only non-empty segments (effectively drops
        # any empty edited segments)
        res_segs	= cur_segs[:len(cur_segs)-len(new_segs)] + list(filter(None, new_segs))
        return '/'.join( res_segs )
    else:
        return edit


class BinanceMainnet( cryptocurrencies.Cryptocurrency ):
    NAME = "Binance"
    SYMBOL = "BSC"
    NETWORK = "mainnet"
    SOURCE_CODE = "https://github.com/bnb-chain/bsc"
    COIN_TYPE = cryptocurrencies.CoinType({
        "INDEX": 60,
        "HARDENED": True
    })

    SCRIPT_ADDRESS = 0x05
    PUBLIC_KEY_ADDRESS = 0x00
    SEGWIT_ADDRESS = cryptocurrencies.SegwitAddress({
        "HRP": "bc",
        "VERSION": 0x00
    })

    EXTENDED_PRIVATE_KEY = cryptocurrencies.ExtendedPrivateKey({
        "P2PKH": 0x0488ade4,
        "P2SH": 0x0488ade4,
        "P2WPKH": 0x04b2430c,
        "P2WPKH_IN_P2SH": 0x049d7878,
        "P2WSH": 0x02aa7a99,
        "P2WSH_IN_P2SH": 0x0295b005
    })
    EXTENDED_PUBLIC_KEY = cryptocurrencies.ExtendedPublicKey({
        "P2PKH": 0x0488b21e,
        "P2SH": 0x0488b21e,
        "P2WPKH": 0x04b24746,
        "P2WPKH_IN_P2SH": 0x049d7cb2,
        "P2WSH": 0x02aa7ed3,
        "P2WSH_IN_P2SH": 0x0295b43f
    })

    MESSAGE_PREFIX = None
    DEFAULT_PATH = f"m/44'/{str(COIN_TYPE)}/0'/0/0"
    WIF_SECRET_KEY = 0x80


class RippleMainnet( cryptocurrencies.Cryptocurrency ):
    """The standard HDWallet.p2pkh_address (Pay to Public Key Hash) encoding is used, w/ a prefix of
    00.  However, the XRP-specific base-58 encoding is used, resulting in a fixed 'r' prefix.

    See: https://xrpl.org/accounts.html#address-encoding.

    """
    NAME = "Ripple"
    SYMBOL = "XRP"
    NETWORK = "mainnet"
    SOURCE_CODE = "https://github.com/ripple/rippled"
    COIN_TYPE = cryptocurrencies.CoinType({
        "INDEX": 144,
        "HARDENED": True
    })

    PUBLIC_KEY_ADDRESS = 0x00  # Results in the prefix r..., when used w/ the Ripple base-58 alphabet
    SEGWIT_ADDRESS = cryptocurrencies.SegwitAddress({
        "HRP": None,
        "VERSION": 0x00
    })

    EXTENDED_PRIVATE_KEY = cryptocurrencies.ExtendedPrivateKey({
        "P2PKH": None,
        "P2SH": None,
        "P2WPKH": None,
        "P2WPKH_IN_P2SH": None,
        "P2WSH": None,
        "P2WSH_IN_P2SH": None,
    })
    EXTENDED_PUBLIC_KEY = cryptocurrencies.ExtendedPublicKey({
        "P2PKH": None,
        "P2SH": None,
        "P2WPKH": None,
        "P2WPKH_IN_P2SH": None,
        "P2WSH": None,
        "P2WSH_IN_P2SH": None,
    })

    MESSAGE_PREFIX = None
    DEFAULT_PATH = f"m/44'/{str(COIN_TYPE)}/0'/0/0"
    WIF_SECRET_KEY = 0x80


class XRPHDWallet( hdwallet.HDWallet ):
    """The XRP address format uses the standard p2pkh_address formulation, from
    https://xrpl.org/accounts.html#creating-accounts:

    The ripemd160 hash of sha256 hash of public key, then base58-encoded w/ 4-byte checksum.  The
    base-58 dictionary used is the standard Ripple (not Bitcoin!) alphabet:

        rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz

    NOTE: Only secp256k1 keypairs are supported; these are the default for the Ripple ledger.

    """
    def p2pkh_address( self ):
        p2pkh_btc	= super( XRPHDWallet, self ).p2pkh_address()
        p2pkh		= base58.b58decode_check( p2pkh_btc )
        return base58.b58encode_check( p2pkh, base58.RIPPLE_ALPHABET ).decode( 'UTF-8' )


class Account:
    """A Cryptocurrency "Account" / Wallet, based on a variety of underlying Python crypto-asset
    support modules.  Presently, only meherett/python-hdwallet is used.

    An appropriate hdwallet-like wrapper is built, for any crypto-asset supported using another
    module.  The required hdwallet API calls are:

      .from_seed	-- start deriving from the provided seed
      .from_mnemonic	-- start deriving from the provided seed via BIP-39/SLIP-39 mnemonic
      .clean_derivation	-- forget any prior derivation path
      .from_path	-- derive a wallet from the specified derivation path
      .p2pkh_address	-- produce a Legacy format address
      .p2sh_address	-- produce a SegWit format address
      .p2wpkh_address	-- produce a Bech32 format address
      .path		-- return the current wallet derivation path
      .private_key	-- return the current wallet's private key

    For testing eg. BIP-38 encrypted wallets:

      .from_private_key	-- import a specific private key
      .from_encrypted	-- import an encrypted wallet

    Also expect the following attributes to be available:

      ._cryptocurrency.SYMBOL:	The short name of the crypto-asset, eg 'XRP'

    Supports producing Legacy addresses for Bitcoin, and Litecoin.  Doge (D...) and Ethereum (0x...)
    addresses use standard BIP44 derivation.

    | Crypto | Semantic | Path              | Address | Support |
    |--------+----------+-------------------+---------+---------|
    | ETH    | Legacy   | m/44'/ 60'/0'/0/0 | 0x...   |         |
    | BSC    | Legacy   | m/44'/ 60'/0'/0/0 | 0x...   | Beta    |
    | BTC    | Legacy   | m/44'/  0'/0'/0/0 | 1...    |         |
    |        | SegWit   | m/49'/  0'/0'/0/0 | 3...    |         |
    |        | Bech32   | m/84'/  0'/0'/0/0 | bc1...  |         |
    | LTC    | Legacy   | m/44'/  2'/0'/0/0 | L...    |         |
    |        | SegWit   | m/49'/  2'/0'/0/0 | M...    |         |
    |        | Bech32   | m/84'/  2'/0'/0/0 | ltc1... |         |
    | DOGE   | Legacy   | m/44'/  3'/0'/0/0 | D...    |         |
    | XRP    | Legacy   | m/44'/144'/0'/0/0 | r...    | Beta    |

    """
    CRYPTO_SYMBOLS		= dict(
        # Convert known Symbols to official Cryptocurrency Name.  By convention, Symbols are
        # capitalized to avoid collisions with names
        ETH		= 'Ethereum',
        BTC		= 'Bitcoin',
        LTC		= 'Litecoin',
        DOGE		= 'Dogecoin',
        BSC		= 'Binance',
        XRP		= 'Ripple',
    )
    CRYPTO_DECIMALS		= dict(
        # Ethereum-related Cryptocurrencies are denominated 10^18, typically default 6 decimals
        # precision.  Bitcoin-related cryptocurrencies are typically 8 decimals precision (1 Sat is
        # 1/10^8 Bitcoin).  For XRP 1 Drop = 1/10^6 Ripple: https://xrpl.org/currency-formats.html
        # For formatting, we'll typically default to decimals//3, which works out fairly well for
        # most Cryptocurrencies and ERC-20 Tokens.  The exception are eg. WBTC (8 decimals) vs. BTC
        # (24 decimals); using the default 8 // 3 = 2 for WBTC would be dramatically too few
        # decimals of precision for practical use.  So, we recommend defaulting to the underlying
        # known cryptocurrency, when a proxy Token is used for price calculations.
        ETH		= 18,
        BTC		= 24,
        LTC		= 24,
        DOGE		= 24,
        BSC		= 18,
        XRP		= 6,
    )
    CRYPTO_NAMES		= dict(
        # Currently supported (in order of visibility), and conversion of known Names to Symbol.  By
        # convention, Cryptocurrency names are lower-cased to avoid collisions with symbols.
        ethereum	= 'ETH',
        bitcoin		= 'BTC',
        litecoin	= 'LTC',
        dogecoin	= 'DOGE',
        binance		= 'BSC',
        ripple		= 'XRP',
    )
    CRYPTOCURRENCIES		= set( CRYPTO_NAMES.values() )
    CRYPTOCURRENCIES_BETA	= set( ('BSC', 'XRP') )

    ETHJS_ENCRYPT		= set( ('ETH', 'BSC') )			# Can be encrypted w/ Ethereum JSON wallet
    BIP38_ENCRYPT		= CRYPTOCURRENCIES - ETHJS_ENCRYPT      # Can be encrypted w/ BIP-38

    CRYPTO_FORMAT		= dict(
        ETH		= "legacy",
        BTC		= "bech32",
        LTC		= "bech32",
        DOGE		= "legacy",
        BSC		= "legacy",
        XRP		= "legacy",
    )

    # Any locally-defined python-hdwallet classes, cryptocurrency definitions, and any that may
    # require some adjustments when calling python-hdwallet address and other functions.
    CRYPTO_WALLET_CLS		= dict(
        XRP		= XRPHDWallet,
    )
    CRYPTO_LOCAL		= dict(
        BSC		= BinanceMainnet,
        XRP		= RippleMainnet,
    )
    CRYPTO_LOCAL_SYMBOL		= dict(
        BSC		= "ETH"
    )

    # The available address formats and default derivation paths.
    FORMATS		= ("legacy", "segwit", "bech32")

    CRYPTO_FORMAT_PATH		= dict(
        ETH		= dict(
            legacy	= "m/44'/60'/0'/0/0",
        ),
        BSC		= dict(
            legacy	= "m/44'/60'/0'/0/0",
        ),
        BTC		= dict(
            legacy	= "m/44'/0'/0'/0/0",
            segwit	= "m/49'/0'/0'/0/0",
            bech32	= "m/84'/0'/0'/0/0",
        ),
        LTC		= dict(
            legacy	= "m/44'/2'/0'/0/0",
            segwit	= "m/49'/2'/0'/0/0",
            bech32	= "m/84'/2'/0'/0/0",
        ),
        DOGE		= dict(
            legacy	= "m/44'/3'/0'/0/0",
        ),
        XRP		= dict(
            legacy	= "m/44'/144'/0'/0/0",
        )
    )

    CRYPTO_FORMAT_SEMANTIC	= dict(
        ETH		= dict(
            legacy	= "p2pkh",
        ),
        BSC		= dict(
            legacy	= "p2pkh",
        ),
        BTC		= dict(
            legacy	= "p2pkh",
            segwit	= "p2wpkh_in_p2sh",
            bech32	= "p2wpkh",
        ),
        LTC		= dict(
            legacy	= "p2pkh",
            segwit	= "p2wpkh_in_p2sh",
            bech32	= "p2wpkh",
        ),
        DOGE		= dict(
            legacy	= "p2pkh",
        ),
        XRP		= dict(
            legacy	= "p2pkh",
        )
    )

    @classmethod
    def path_default( cls, crypto, format=None ):
        """Return the default derivation path for the given crypto, based on its currently selected default
        address format.

        """
        crypto			= cls.supported( crypto )
        format			= format.lower() if format else cls.address_format( crypto )
        if format not in cls.CRYPTO_FORMAT_PATH[crypto]:
            raise ValueError( f"{format} not supported for {crypto}; specify one of {commas( cls.CRYPTO_FORMAT_PATH[crypto].keys() )}" )
        return cls.CRYPTO_FORMAT_PATH[crypto][format]

    @classmethod
    def address_format( cls, crypto, format=None ):
        """Get or set the desired default address format for the specified supported crypto.  Future
        instances of Address created for the crypto will use the specified address format and its
        default derivation path.

        """
        crypto			= cls.supported( crypto )
        if format is None:
            return cls.CRYPTO_FORMAT[crypto]

        format			= format.lower() if format else None
        if format not in cls.FORMATS:
            raise ValueError( f"{crypto} address format {format!r} not recognized; specify one of {commas( cls.FORMATS )}" )
        cls.CRYPTO_FORMAT[crypto]	= format

    @classmethod
    def supported( cls, crypto ):
        """Validates that the specified cryptocurrency is supported and returns the normalized "SYMBOL"
        for it, or raises an a ValueError.  Eg. "ETH"/"Ethereum" --> "ETH"

        """
        try:
            validated		= cls.CRYPTO_NAMES.get(
                crypto.lower(),
                crypto.upper() if crypto.upper() in cls.CRYPTO_SYMBOLS else None
            )
            if validated:
                return validated
        except Exception as exc:
            validated		= exc
            raise
        finally:
            log.debug( f"Validating {crypto!r} yields: {validated!r}" )

        raise ValueError( f"{crypto} not presently supported; specify {commas( cls.CRYPTOCURRENCIES )}" )

    def __str__( self ):
        """Until from_seed/from_path are invoked, may not have an address or derivation path."""
        address			= None
        try:
            address		= self.address
        except Exception:
            pass
        return f"{self.crypto}: {address}"

    def __repr__( self ):
        return f"{self.__class__.__name__}({self} @{self.path})"

    def __init__( self, crypto, format=None ):
        crypto			= Account.supported( crypto )
        cryptocurrency		= self.CRYPTO_LOCAL.get( crypto )
        self.format		= format.lower() if format else Account.address_format( crypto )
        semantic		= self.CRYPTO_FORMAT_SEMANTIC[crypto][self.format]
        hdwallet_cls		= self.CRYPTO_WALLET_CLS.get( crypto, hdwallet.HDWallet )
        if hdwallet_cls is None:
            raise ValueError( f"{crypto} does not support address format {self.format}" )
        self.hdwallet		= hdwallet_cls( symbol=crypto, cryptocurrency=cryptocurrency, semantic=semantic )

    def from_seed( self, seed: Union[str,bytes], path: Optional[str] = None, format=None ) -> Account:
        """Derive the Account from the supplied seed and (optionally) path; uses the default derivation
        path for the Account address format, if None provided.  As with all of the functions that
        completely replace the derivation Seed, we clear any existing known derivation path; it is
        unrelated to this newly supplied seed.  Handles bytes or hex seeds, optionally with "0x...".

        """
        if type( seed ) is bytes:
            seed		= codecs.encode( seed, 'hex_codec' ).decode( 'ascii' )
        if seed[:2].lower() == "0x":
            seed		= seed[2:]
        assert all( c in string.hexdigits for c in seed ), \
            "Only bytes and hex string HD Wallet Seeds are supported"

        self.hdwallet.clean_derivation()
        self.hdwallet.from_seed( seed )
        self.from_path( path )
        return self

    def from_mnemonic( self, mnemonic: str, path: Optional[str] = None, passphrase: Optional[Union[bytes,str]] = None, using_bip39: bool = False ) -> Account:
        """Derive the Account seed from the supplied BIP-39/SLIP-39 Mnemonic(s) and (optionally) path.
        Since Mnemonics are intended to encode "root" HD Wallet seeds, uses the default derivation
        path for the Account address format, if None provided.

        SLIP-39 Mnemonics are recognized by the fact that they are multiple lines (has newlines).

        If 'using_bip39', then any supplied SLIP-39 Mnemonics entropy will first be converted back
        into a BIP-39 Mnemonic (to maintain compatibility w/ BIP-39 wallets) to obtain the Seed.
        Otherwise, SLIP-39 Mnemonics will use native SLIP-39 Seed decoding.

        """
        mnemonics_lines		= [
            s.strip()
            for s in mnemonic.split( '\n' )
            if s.strip()
        ]
        if len( mnemonics_lines ) < 1:
            raise ValueError( "At least one BIP-39 2 SLIP-39 Mnemonics required" )
        if len( mnemonics_lines ) > 1:
            # Must be SLIP-39 Mnemonic Phrases
            seed		= recover_slip39( mnemonics_lines, passphrase=passphrase, using_bip39=using_bip39 )
            return self.from_seed( seed=seed, path=path )
        # Must be a single BIP-39 Mnemonic Phrase (as a UTF-8 string)
        if isinstance( passphrase, bytes ):
            passphrase		= passphrase.decode( 'UTF-8' )

        self.hdwallet.clean_derivation()
        self.hdwallet.from_mnemonic( *mnemonics_lines, passphrase=passphrase )  # python-hdwallet requires str/None
        self.from_path( path )
        return self

    def from_xpubkey( self, xpubkey: str, path: Optional[str] = None ) -> Account:
        """Derive the Account from the supplied xpubkey and (optionally) path; uses no derivation path
        by default derivation path for the Account address format, if None provided.

        Since this xpubkey may have been generated at some arbitrary path, eg.

            m/44'/60'/0'

        any subsequent path provided here, such as "m/0/0" will be "added" to the original
        derivation path, to give us the address at effective path eg.:

            m/44'/60'/0'/0/0

        However, if we ask for the self.path from this account, it will return only the portion
        provided here:

            m/0/0

        It is impossible to correctly recover any "hardened" accounts from an xpubkey, such as
        "m/1'/0".  These would need access to the private key material, which is missing.
        Therefore, the original account (or an xprivkey) would be required to access the desired
        path:

            m/44'/60'/0'/1'/0

        """
        self.hdwallet.clean_derivation()
        self.hdwallet.from_xpublic_key( xpubkey )
        self.from_path( path or "m/" )
        return self

    def from_xprvkey( self, xprvkey: str, path: Optional[str] = None ) -> Account:
        self.hdwallet.clean_derivation()
        self.hdwallet.from_xprivate_key( xprvkey )
        self.from_path( path or "m/" )
        return self

    def from_public_key( self, public_key: str, path: Optional[str] = None ) -> Account:
        self.hdwallet.clean_derivation()
        self.hdwallet.from_public_key( public_key )
        self.from_path( path or "m/" )
        return self

    def from_private_key( self, private_key: str, path: Optional[str] = None ) -> Account:
        self.hdwallet.clean_derivation()
        self.hdwallet.from_private_key( private_key )
        self.from_path( path or "m/" )
        return self

    def from_path( self, path: Optional[str] = None ) -> Account:
        """Change the Account to derive from the provided path (or from the default path, if currently
        empty, ie. 'm/').

        If a partial path is provided (eg "...1'/0/3"), then use it to replace the given segments in
        current (or the default) account path, leaving the remainder alone.

        If the derivation path is empty (only "m/") then leave the Account at clean_derivation state

        """
        from_path		= self.path
        log.debug( f"Changing {self.format} {self!r} from {from_path} to {path}" )
        if not from_path or len( from_path ) <= 2:
            if from_path != "m/":
                raise ValueError( f"Empty but invalid path detected: {from_path}" )
            log.debug( f"Default path for {self}, was {from_path!r}" )
            from_path		= Account.path_default( self.crypto, self.format )
        if path:
            into_path		= path_edit( from_path, path )
            log.debug( f"Editing path for {self}, from {from_path!r} w/ {path!r}, into {into_path!r}" )
            from_path		= into_path
        # Valid HD wallet derivation paths always start with "m/"
        if not ( from_path and len( from_path ) >= 2 and from_path.startswith( "m/" ) ):
            raise ValueError( f"Unrecognized HD wallet derivation path: {from_path!r}" )
        if len( from_path ) > 2:
            self.hdwallet.from_path( from_path )
        return self

    @property
    def address( self ):
        """Returns the 1..., 3... or bc1... address, depending on whether format is legacy, segwit or bech32"""
        return self.formatted_address()

    def formatted_address( self, format=None ):
        if ( format or self.format or '' ).lower() == "legacy":
            return self.legacy_address()
        elif ( format or self.format or '' ).lower() == "segwit":
            return self.segwit_address()
        elif ( format or self.format or '' ).lower() == "bech32":
            return self.bech32_address()
        raise ValueError( f"Unknown addresses semantic: {self.format}" )

    def substitute_symbol( method ):
        """For some locally-defined cryptocurrencies, the python-hdwallet code specifically checks
        for "ETH" and represents the address in the standard Ethereum 0x... format.  For those
        cryptocurrencies that need the underlying hdwallet._cryptocurrency.symbol adjusted during
        the call, place the entry in the CRYPTO_LOCAL_SYMBOL dict, above.

        Decorate the target methods that require the underlying .hdwallet._cryptocurrency.SYMBOL to
        be adjusted for the duration of the method.

        See: python-hdwallet/hdwallet/hdwallet.py, line 1102.

        """
        @wraps( method )
        def wrapper( self, *args, **kwds ):
            symbol              = self.hdwallet._cryptocurrency.SYMBOL
            try:
                self.hdwallet._cryptocurrency.SYMBOL \
                                = self.CRYPTO_LOCAL_SYMBOL.get( symbol, symbol )        # noqa: E126
                return method( self, *args, **kwds )
            finally:
                self.hdwallet._cryptocurrency.SYMBOL \
                                = symbol						# noqa: E126
        return wrapper

    @substitute_symbol
    def legacy_address( self ):
        """BIP-44 Address"""
        return self.hdwallet.p2pkh_address()

    def segwit_address( self ):
        """BIP-49 Address"""
        return self.hdwallet.p2wpkh_in_p2sh_address()

    def bech32_address( self ):
        """BIP-84 Address"""
        return self.hdwallet.p2wpkh_address()

    @property
    def name( self ):
        return self.hdwallet._cryptocurrency.NAME

    @property
    def symbol( self ):
        return self.hdwallet._cryptocurrency.SYMBOL
    crypto		= symbol

    @property
    def path( self ) -> str:
        return self.hdwallet.path() or 'm/'

    @property
    def key( self ):
        return self.hdwallet.private_key()
    prvkey		= key

    @property
    def xkey( self ):
        return self.hdwallet.xprivate_key()
    xprvkey		= xkey

    @property
    def pubkey( self ):
        return self.hdwallet.public_key()

    @property
    def xpubkey( self ):
        """Returns the xpub, ypub or zpub, depending on whether format is legacy, segwit or bech32.
        The HD wallet account represented by this xpub... key is that of the current derivation
        path, eg. "m/44'/60'/0'/0/0" for the default ETH wallet.  Thus, when restoring using
        eg. from_xpubkey, the default path used should be empty, ie. "m/".

        """
        return self.hdwallet.xpublic_key()

    def encrypted( self, passphrase ):
        """Output the appropriately encrypted private key for this cryptocurrency.  Ethereum uses
        encrypted JSON wallet standard, Bitcoin et.al. use BIP-38 encrypted private keys.

        A BIP-39 encrypted wallet encodes the private key at a certain derivation path, which is not
        remembered in the BIP-39 encoding!  Therefore, when you recover a BIP-39 encrypted wallet,
        you must tell the Account what derivation path you want to derive.  The default for None is
        NO derivation path (ie. "m/") -- the exact same wallet at whatever the derivation path was
        when .encrypted was invoked, is what will be recovered.

        """
        if self.crypto in self.ETHJS_ENCRYPT:
            if not eth_account:
                raise NotImplementedError( "The eth-account module is required to support Ethereum JSON wallet encryption; pip install slip39[wallet]" )
            wallet_dict		= eth_account.Account.encrypt( self.key, passphrase )
            return json.dumps( wallet_dict, separators=(',',':') )
        return self.bip38( passphrase )

    def from_encrypted( self, encrypted_privkey, passphrase, strict=True, path: Optional[str] = None ):
        """Import the appropriately decrypted private key for this cryptocurrency."""
        if self.crypto in self.ETHJS_ENCRYPT:
            if not eth_account:
                raise NotImplementedError( "The eth-account module is required to support Ethereum JSON wallet decryption; pip install slip39[wallet]" )
            private_hex		= bytes( eth_account.Account.decrypt( encrypted_privkey, passphrase )).hex()
            self.from_private_key( private_hex, path=path )
            return self
        return self.from_bip38( encrypted_privkey, passphrase=passphrase, strict=strict )

    def bip38( self, passphrase, flagbyte=b'\xe0' ):
        """BIP-38 encrypt the private key"""
        if not scrypt or not AES:
            raise NotImplementedError( "The scrypt module is required to support BIP-38 encryption; pip install slip39[wallet]" )
        if self.crypto not in self.BIP38_ENCRYPT:
            raise NotImplementedError( f"{self.crypto} does not support BIP-38 private key encryption" )
        private_hex		= self.key
        addr			= self.legacy_address().encode( 'UTF-8' )  # Eg. b"184xW5g..."
        ahash			= hashlib.sha256( hashlib.sha256( addr ).digest() ).digest()[0:4]
        if isinstance( passphrase, str ):
            passphrase		= passphrase.encode( 'UTF-8' )
        key			= scrypt( passphrase or b"", salt=ahash, key_len=64, N=16384, r=8, p=8 )
        derivedhalf1		= key[0:32]
        derivedhalf2		= key[32:64]
        aes			= AES.new( derivedhalf2, AES.MODE_ECB )
        enchalf1		= aes.encrypt( ( int( private_hex[ 0:32], 16 ) ^ int.from_bytes( derivedhalf1[ 0:16], 'big' )).to_bytes( 16, 'big' ))
        enchalf2		= aes.encrypt( ( int( private_hex[32:64], 16 ) ^ int.from_bytes( derivedhalf1[16:32], 'big' )).to_bytes( 16, 'big' ))
        prefix			= b'\x01\x42'
        encrypted_privkey	= prefix + flagbyte + ahash + enchalf1 + enchalf2
        # Encode the encrypted private key to base58, adding the 4-byte base58 check suffix
        return base58.b58encode_check( encrypted_privkey ).decode( 'UTF-8' )

    def from_bip38( self, encrypted_privkey, passphrase, path: Optional[str] = None, strict: bool = True ):
        """Bip-38 decrypt and import the private key."""
        if not scrypt or not AES:
            raise NotImplementedError( "The scrypt module is required to support BIP-38 decryption; pip install slip39[wallet]" )
        if self.crypto not in self.BIP38_ENCRYPT:
            raise NotImplementedError( f"{self.crypto} does not support BIP-38 private key decryption" )
        # Decode the encrypted private key from base58, discarding the 4-byte base58 check suffix
        d			= base58.b58decode_check( encrypted_privkey )
        assert len( d ) == 43 - 4, \
            f"BIP-38 encrypted key should be 43 bytes long, not {len( d ) + 4} bytes"
        pre,flag,ahash,eh1,eh2	= d[0:2],d[2:3],d[3:7],d[7:7+16],d[7+16:7+32]
        assert pre == b'\x01\x42', \
            f"Unrecognized BIP-38 encryption prefix: {pre!r}"
        assert flag in ( b'\xc0', b'\xe0' ), \
            f"Unrecognized BIP-38 flagbyte: {flag!r}"
        if isinstance( passphrase, str ):
            passphrase		= passphrase.encode( 'UTF-8' )
        key			= scrypt( passphrase or b"", salt=ahash, key_len=64, N=16384, r=8, p=8 )
        derivedhalf1		= key[0:32]
        derivedhalf2		= key[32:64]
        aes			= AES.new( derivedhalf2, AES.MODE_ECB )
        dechalf2		= aes.decrypt( eh2 )
        dechalf1		= aes.decrypt( eh1 )
        priv			= dechalf1 + dechalf2
        priv			= ( int.from_bytes( priv, 'big' ) ^ int.from_bytes( derivedhalf1, 'big' )).to_bytes( 32, 'big' )
        # OK, we have the Private Key; we can recover the account, then verify the
        # remainder of the checks
        private_hex		= codecs.encode( priv, 'hex_codec' ).decode( 'ascii' )
        self.from_private_key( private_hex, path=path )
        addr			= self.legacy_address().encode( 'UTF-8' )  # Eg. b"184xW5g..."
        ahash_confirm		= hashlib.sha256( hashlib.sha256( addr ).digest() ).digest()[0:4]
        if ahash_confirm != ahash:
            warning		= f"BIP-38 address hash verification failed ({ahash_confirm.hex()} != {ahash.hex()}); passphrase may be incorrect."
            if strict:
                raise AssertionError( warning )
            else:
                log.warning( warning )
        return self


def path_parser(
    paths: str,
    allow_unbounded: bool	= True,
) -> Tuple[str, Dict[str, Callable[[], int]]]:
    """Create a format and a dictionary of iterators to feed into it.

    Supports paths with an arbitrary prefix, eg. 'm/' or '../'
    """
    path_segs			= paths.split( '/' )
    unbounded			= False
    ranges			= {}

    for i,s in list( enumerate( path_segs )):
        if '-' not in s:
            continue
        c			= chr(ord('a')+i)
        tic			= s.endswith( "'" )
        if tic:
            s			= s[:-1]
        b,e			= s.split( '-' )
        b			= int( b or 0 )
        if e:
            e			= int( e )
            ranges[c]		= lambda b=b,e=e: range( b, e+1 )
        else:
            assert allow_unbounded and not ( unbounded or ranges ), \
                f"{'Only first' if allow_unbounded else 'No'} range allowed to be unbounded;" \
                f" this is the {ordinal(len(ranges)+1)} range in {paths}"
            unbounded		= True
            ranges[c]		= lambda b=b: itertools.count( b )
        path_segs[i]		= f"{{{c}}}" + ( "'" if tic else "" )

    path_fmt			= '/'.join( path_segs )
    return path_fmt, ranges


def path_sequence(
    path_fmt: str,
    ranges: Dict[str, Callable[[], int]],
):
    """Yield a sequence of paths, modulating the format specifiers of the
    path_fmt according to their value sources in ranges.

    For example, a

        path_fmt = "m/44'/60'/0'/0/{f}", with a
        ranges   = dict( f=lambda b=0, e=2: range( b, e+1 ) )

    would yield the paths:

        "m/44'/60'/0'/0/0"
        "m/44'/60'/0'/0/1"
        "m/44'/60'/0'/0/2"
    """
    # Stert all the iterators
    viters			= {
        k: iter( l() )
        for k,l in ranges.items()
    }
    values			= {		# Initial round of values; must provide at least one
        k: next( viters[k] )
        for k in viters
    }
    assert all( v is not None for v in values.values() ), \
        "Iterators for path segment values must yield at least an initial value"

    while not any( v is None for v in values.values() ):
        yield path_fmt.format( **values )
        if not ranges:
            break				# No variable records at all; just one
        # Get the next value.  Working from the lowest iterator up, cycle value(s)
        for i,k in enumerate( sorted( viters.keys(), reverse=True )):
            values[k]		= next( viters[k], None )
            if values[k] is not None:
                break
            # OK, this iterable has ended.  Restart it, and cycle to the next one up, iff
            # there are remaining ranges
            if i+1 < len( ranges ):
                viters[k]	= iter( ranges[k]() )
                values[k]	= next( viters[k], None )


def path_hardened( path ):
    """Remove any non-hardened components from the end of path, eg:

    >>> path_hardened( "m/84'/0'/0'/1/2" )
    ("m/84'/0'/0'", 'm/1/2')
    >>> path_hardened( "m/1" )
    ('m/', 'm/1')
    >>> path_hardened( "m/1'" )
    ("m/1'", 'm/')
    >>> path_hardened( "m/" )
    ('m/', 'm/')
    >>> path_hardened( "m/1/2/3'/4" )
    ("m/1/2/3'", 'm/4')

    Returns  the two components as a tuple of two paths
    """
    segs			= path.split( '/' )
    # Always leaves the m/ on the hard path
    for hardened in range( 1, len( segs ) + 1 ):
        if not any( "'" in s for s in segs[hardened:] ):
            break
    else:
        log.debug( f"No non-hardened path segments in {path}" )

    hard			= 'm/' + '/'.join( segs[1:hardened] )
    soft			= 'm/' + '/'.join( segs[hardened:] )
    return hard,soft


def cryptopaths_parser(
    cryptocurrency,  # Or a cryptopaths list, eg. ["XRP",(ETH,"m/.."),(BTC,NONE,"segwit")]
    edit			= None,
    hardened_defaults		= False,
    format			= None,
):
    """Generate a standard cryptopaths list, from the given sequnce of "<crypto>",
    (<crypto>,<paths>), (<crypto>,<paths>,<format>), or "<crypto>[:<paths>[:<format>:]"
    cryptocurrencies (default: CRYPTO_PATHS, optionally w/ only the hardened portion of the path,
    eg. omitting the trailing ../0/0).

    Adjusts the provided derivation paths by an optional eg. "../-" path adjustment.

    A non-default format may be specified, which may change the default HD derivation path.  This
    must also be passed back, as it also affects the crypto's account's address format.

    """
    for crypto in cryptocurrency or CRYPTO_PATHS:
        if type(crypto) is str:
            crypto		= crypto.split( ':' )

        cry,*pth		= crypto
        pth,*fmt		= pth or (None,)
        fmt,			= fmt or (format,)

        cry			= Account.supported( cry )
        if not pth:
            pth			= Account.path_default( cry, fmt )
            if hardened_defaults:
                pth,_		= path_hardened( pth )
        if edit:
            pth			= path_edit( pth, edit )
        yield (cry,pth,fmt)


def random_secret(
    seed_length: Optional[int]
) -> bytes:
    """Generates a new random secret.

    NOTE: There is a slightly less than 1 / 2^128 chance that any given random secret will lead to
    an invalid BTC wallet private key!  This is because the 256-bit seed for bitcoin must be less than
    the secp256k1 field size:

        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141

    We cannot generate secrets that are guaranteed to be valid for every derived HD Wallet BTC
    address.  The secret is derived through a complex hashing procedure for each wallet at each
    path.  The probability of this occurring is so vanishingly small that we will instead simply opt
    to *not* generate an public wallet address, when asked.

    Just to give you an understandable reference, the first 128 bits of an illegal secret unusable
    for generating a Bitcoin wallet must have 128 '1' bits at its start.  If wallets were generated
    at a billion per second, by every person alive on earth, it should take about 1.5 trillion years
    to arrive at the first invalid seed.

    2**128 / 1e9 / 7e9 / (60*60*24*365) == 1,541,469,010,115.145

    """
    assert seed_length, \
        f"Must supply a non-zero length in bytes, not {seed_length}"
    return RANDOM_BYTES( seed_length )


Details = namedtuple( 'Details', ('name', 'group_threshold', 'groups', 'accounts', 'using_bip39') )


def enumerate_mnemonic( mnemonic ):
    """Return a dict containing the supplied mnemonics stored by their indexed, starting from 0.
    Each Mnemonic is labelled with its ordinal index (ie. beginning at 1).

    """
    if isinstance( mnemonic, str ):
        mnemonic		= mnemonic.split( ' ' )
    return dict(
        (i, f"{i+1:>2d} {w}")
        for i,w in enumerate( mnemonic )
    )


def organize_mnemonic( mnemonic, rows=None, cols=None, label="" ):
    """Given a SLIP-39 "word word ... word" or ["word", "word", ..., "word"] mnemonic, emit rows
    organized in the desired rows and cols (with defaults, if not provided).  We return the fully
    formatted line, plus the list of individual words in that line.

    """
    num_words			= enumerate_mnemonic( mnemonic )
    if not rows or not cols:
        rows,cols		= MNEM_ROWS_COLS.get( len(num_words), (7, 3))
    for r in range( rows ):
        line			= label if r == 0 else ' ' * len( label )
        words			= []
        for c in range( cols ):
            word		= num_words.get( c * rows + r )
            if word:
                words.append( word )
                line	       += f"{word:<13}"
        yield line,words


def group_parser( group_spec ):
    """Parse a SLIP-39 group specification.

        Frens6, Frens 6, Frens(6)	- A 3/6 group (default is 1/2 of group size, rounded up)
        Frens2/6, Frens(2/6)		- A 2/6 group

    """
    match			= group_parser.RE.match( group_spec )
    if not match:
        raise ValueError( f"Invalid group specification: {group_spec!r}" )
    name			= match.group( 'name' )
    size			= match.group( 'size' )
    require			= match.group( 'require' )
    if not size:
        size			= 1
    if not require:
        # eg. default 2/4, 3/5
        require			= math.ceil( int( size ) * GROUP_REQUIRED_RATIO )
    return name,(int(require),int(size))
group_parser.RE			= re.compile( # noqa E305
    r"""^
        \s*
        (?P<name> [^\d\(/]+ )
        \s*\(?\s*
        (?: (?P<require> \d* ) \s* / )?
        \s*
        (?P<size> \d* )
        \s*\)?\s*
        $""", re.VERBOSE )


def create(
    name: str,
    group_threshold: Optional[Union[int,float]] = None,		# Default: 1/2 of groups, rounded up
    groups: Optional[Union[List[str],Dict[str,Tuple[int, int]]]] = None,  # Default: 4 groups (see defaults.py)
    master_secret: Optional[Union[str,bytes]] = None,		# Default: generate 128-bit Seed Entropy
    passphrase: Optional[Union[bytes,str]] = None,
    using_bip39: Optional[bool]	= None,  # Produce wallet Seed from master_secret Entropy using BIP-39 generation
    iteration_exponent: int	= 1,
    cryptopaths: Optional[Sequence[Union[str,Tuple[str,str],Tuple[str,str,str]]]] = None,  # default: ETH, BTC at default path, format
    strength: Optional[int]	= None,		# Default: 128
) -> Tuple[str,int,Dict[str,Tuple[int,List[str]]], Sequence[Sequence[Account]], bool]:
    """Creates a SLIP-39 encoding for supplied master_secret Entropy, and 1 or more Cryptocurrency
    accounts.  Returns the Details, in a form directly compatible with the layout.produce_pdf API.

    The master_secret Seed Entropy is discarded (because it is, of course, always recoverable from
    the SLIP-39 mnemonics).

    We strive to default to "do the right thing", here.  If you supply BIP-39 Mnemonics, we'll
    round-trip them via SLIP-39, and produce compatible crypto account addresses.  This should be
    the "typical" case, as most people already have BIP-39 Mnemonics.  If you don't have a BIP-39
    Mnemonic, make sure you set using_bip39 = True; this *also* implies that you *MUST* have already
    converted your entropy to BIP-39 (or are going to recover it and do so, later), so we'll warn
    you to do that.  If a passphrase is provided, it is assumed to be a BIP-39 passphrase, and is used
    to generate the crypto account addresses -- it will *not* be used to "encrypt" the SLIP-39!

    If you supply raw entropy, we'll assume you have SLIP-39 compatible wallet and want to use it
    directly.

    Creates accountgroups derived from the Seed Entropy.  By default, this is done in the SLIP-39
    standard, using the master_secret Entropy directly.  If a passphrase is supplied, this is also
    used in the SLIP-39 standard fashion (not recommended -- not Trezor "Model T" compatible).

    If using_bip39, creates the Cryptocurrency accountgroups from the supplied master_secret
    Entropy, by generating the Seed from a BIP-38 Mnemonic produced from the provided entropy
    (or generated, default 128 bits), plus any supplied passphrase.

    """
    if master_secret is None:
        if not strength:
            strength		= BITS_DEFAULT
        master_secret		= random_secret( strength // 8 )
    if isinstance( master_secret, bytes ) and not ( 128 <= len( master_secret ) * 8 <= 512 ):
        log.warning( f"Strangely sized {len(master_secret) * 8}-bit entropy provided; may be weak" )

    # If a non-hex str is passed as entropy, assume it is a BIP-39 Mnemonic, and that we want to use
    # SLIP-39 to round-trip the underlying BIP-39 entropy AND derive compatible wallets.
    if isinstance( master_secret, str ) and all( c in '0123456789abcdef' for c in master_secret.lower() ):
        master_secret		= codecs.decode( master_secret, 'hex_codec' )
    if using_bip39 is None and isinstance( master_secret, str ):
	# Assume it must be a BIP-39 Mnemonic
        using_bip39		= True
    else:
        # Assume caller knows; default False (use SLIP-39 directly)
        using_bip39		= bool( using_bip39 )

    # Derive the desired account(s) at the specified derivation paths, or the default, using either
    # BIP-39 Seed generation, or directly from Entropy for SLIP-39.
    if using_bip39:
        # For BIP-39, the passphrase is consumed here, and Cryptocurrency accounts are generated
        # using the BIP-39 Seed generated from entropy + passphrase.  This should be the "typical"
        # use-case, where someone already has a BIP-39 Mnemonic and/or wants to use a "standard"
        # BIP-39 compatible hardware wallet.
        log.warning( f"Assuming BIP-39 seed entropy: Ensure you recover and use via a BIP-39 Mnemonic" )
        if isinstance( master_secret, str ):
            master_secret	= recover_bip39( mnemonic=master_secret, as_entropy=True )
        bip39_mnem		= produce_bip39( entropy=master_secret )
        bip39_seed		= recover_bip39(
            mnemonic	= bip39_mnem,
            passphrase	= passphrase,
        )
        log.info(
            f"SLIP-39 for {name} from {len(master_secret)*8}-bit Entropy using BIP-39 Mnemonic{' w/ Passphrase' if passphrase else ''}"
        )
        accts			= list( accountgroups(
            master_secret	= bip39_seed,
            cryptopaths		= cryptopaths,
            allow_unbounded	= False,
        ))
        passphrase		= None  # Consumed by BIP-39; not used for SLIP-39 Mnemonics "backup"!
    else:
        # For SLIP-39, accounts are generated directly from supplied Entropy, and passphrase
        # encrypts the SLIP-39 Mnemonics, below.  Using a SLIP-39 with a passphrase is so unlikely
        # to be correct that we will warn about it!  You almost *always* want to use SLIP-39
        # *without* a passphase; use eg. Trezor "hidden wallets" instead.
        (log.warning if passphrase else log.info)(
            f"SLIP-39 for {name} from {len(master_secret)*8}-bit Entropy directly{' w/ SLIP-39 Passphrase' if passphrase else ''}"
        )
        accts			= list( accountgroups(
            master_secret	= master_secret,
            cryptopaths		= cryptopaths,
            allow_unbounded	= False,
        ))

    # Deduce groups, using defaults
    if not groups:
        groups			= GROUPS
    if not is_mapping( groups ):
        if isinstance( groups, str ):
            groups		= groups.split( "," )
        groups			= dict( map( group_parser, groups ))
    g_names,g_dims		= list( zip( *groups.items() ))

    # Generate the SLIP-39 Mnemonics representing the supplied master_secret Seed Entropy.  This
    # always recovers the Seed Entropy; if not using_bip39, this is also the wallet derivation Seed;
    # if using_bip39, the wallet derivation Seed was produced from the BIP-39 Seed generation
    # process (which consumes any passphrase, and the SLIP-39 passphrase is always None, here).
    mnems			= mnemonics(
        group_threshold	= group_threshold,
        groups		= g_dims,
        master_secret	= master_secret,
        passphrase	= passphrase,
        iteration_exponent= iteration_exponent
    )

    groups			= {
        g_name: (g_of, g_mnems)
        for (g_name,(g_of, _),g_mnems) in zip( g_names, g_dims, mnems )
    }
    if log.isEnabledFor( logging.INFO ):
        group_reqs			= list(
            f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
            for g_nam,(g_of,g_mns) in groups.items() )
        requires		= f"Recover w/ {group_threshold} of {len(groups)} groups {commas( group_reqs )}"
        for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
            log.info( f"{g_name}({g_of}/{len(g_mnems)}): {requires}" )
            for mn_n,mnem in enumerate( g_mnems ):
                for line,_ in organize_mnemonic( mnem, label=f"{ordinal(mn_n+1)} " ):
                    log.info( f"{line}" )

    return Details(name, group_threshold, groups, accts, using_bip39)


def mnemonics(
    group_threshold: Optional[int],  # Default: 1/2 of groups, rounded up
    groups: Sequence[Tuple[int, int]],
    master_secret: Optional[Union[str,bytes]] = None,
    passphrase: Optional[Union[bytes,str]] = None,
    iteration_exponent: int	= 1,
    strength: int		= BITS_DEFAULT,
) -> List[List[str]]:
    """Generate SLIP39 mnemonics for the supplied group_threshold of the given groups.  Will generate a
     random master_secret, if necessary.

    If you have BIP-39/SLIP-39 Mnemonic(s), use recovery.recover or .recover_bip39 first.  To
    "backup" a BIP-39 Mnemonic Phrase, you probably want to use .recover_bip39( ..., as_entropy=True
    ) to get the original 128- or 256-bit Entropy, and then produce SLIP-39 Mnemonics from that.

    Later, supply these SLIP-39 Mnemonics to any of the .account... functions with using_bip39=True,
    to derive the original BIP-39 wallets.

    """
    if master_secret is None:
        assert strength in BITS, f"Invalid {strength}-bit secret length specified"
        master_secret		= random_secret( strength // 8 )
    if len( master_secret ) * 8 not in BITS:
        raise ValueError(
            f"Only {commas( BITS, final='and' )}-bit seeds supported; {len(master_secret)*8}-bit seed supplied" )
    if isinstance( passphrase, str ):
        passphrase		= passphrase.encode( 'UTF-8' )

    groups			= list( groups )
    if not group_threshold:
        group_threshold		= math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )
    return generate_mnemonics(
        group_threshold	= group_threshold,
        groups		= groups,
        master_secret	= master_secret,
        passphrase	= passphrase or b"",  # python-shamir-mnemonic requires bytes
        iteration_exponent = iteration_exponent,
    )


def account(
    master_secret: Union[str,bytes],
    crypto: Optional[str]	= None,  # default 'ETH'
    path: Optional[str]		= None,  # default to the crypto's path_default
    format: Optional[str]	= None,  # eg. 'bech32', or use the default address format for the crypto
    passphrase: Optional[Union[bytes,str]] = None,  # If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
):
    """Generate an HD wallet Account from the supplied master_secret seed, at the given HD derivation
    path, for the specified cryptocurrency.

    If the master_secret is bytes, it is used as-is.  If a str, then we generally expect it to be
    hex.  However, this is where we can detect alternatives like "{x,y,z}{pub,priv}key...".  These
    are identifiable by their prefix, which is incompatible with hex, so there is no ambiguity.

    """
    if isinstance( master_secret, str ):
        master_secret		= master_secret.strip()
    if isinstance( master_secret, bytes ) or master_secret[:2].lower == "0x" or all(
        c in string.hexdigits for c in master_secret
    ):
        # Probably a binary/hex Seed.
        acct			= Account(
            crypto	= crypto or 'ETH',
            format	= format,
        )
        acct.from_seed(
            seed	= master_secret,
            path	= path,
        )
        log.debug( f"Created {acct.format} {acct} from {len(master_secret)*8}-bit seed, at derivation path {acct.path}" )
    elif ' ' in master_secret:
        # Some kind of Mnemonic; this is the only valid use of whitespace within a master_secret.
        acct			= Account(
            crypto	= crypto or 'ETH',
            format	= format,
        )
        log.debug( f"Making  {acct.format} {acct} from Mnemonic(s), at derivation path {acct.path}" )
        acct.from_mnemonic( master_secret, path=path, passphrase=passphrase, using_bip39=using_bip39 )
        log.debug( f"Created {acct.format} {acct} from Mnemonic(s), at derivation path {acct.path}" )
    else:
        # See if we recognize the prefix as a {x,y,z}pub... or .prv...  Get the bound function for
        # initializing the seed.  Also, deduce the default format from the x/y/z+pub/prv.
        default_fmt,from_method	= {
            'xpub': ('legacy', Account.from_xpubkey),
            'xprv': ('legacy', Account.from_xprvkey),
            'ypub': ('segwit', Account.from_xpubkey),
            'yprv': ('segwit', Account.from_xprvkey),
            'zpub': ('bech32', Account.from_xpubkey),
            'zprv': ('bech32', Account.from_xprvkey),
        }.get( master_secret[:4], (None,None) )
        if from_method is None:
            raise ValueError(
                f"Only x/y/z + pub/prv prefixes supported; {master_secret[:8]+'...'!r} prefix supplied" )
        if format is None:
            format		= default_fmt
        acct			= Account(
            crypto	= crypto or 'ETH',
            format	= format
        )
        from_method( acct, master_secret, path )  # It's an unbound method, so pass the instance
        log.debug( f"Created {acct.format} {acct} from {master_secret[:4]} key, at derivation path {acct.path}" )

    return acct


def accounts(
    master_secret: Union[str,bytes],
    crypto: str			= None,  # default 'ETH'
    paths: str			= None,  # default to the crypto's path_default; allow ranges
    format: Optional[str]	= None,
    allow_unbounded		= True,
    passphrase: Optional[Union[bytes,str]] = None,  # If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
):
    """Create accounts for crypto, at the provided paths (allowing ranges), with the optionsal address format. """
    for path in [None] if paths is None else path_sequence( *path_parser(
        paths		= paths,
        allow_unbounded	= allow_unbounded,
    )):
        yield account(
            master_secret,
            crypto	= crypto,
            path	= path,
            format	= format,
            passphrase	= passphrase,
            using_bip39	= using_bip39,
        )


def accountgroups(
    master_secret: Union[str,bytes],
    cryptopaths: Optional[Sequence[Union[str,Tuple[str,str],Tuple[str,str,str]]]] = None,  # default: ETH, BTC at default path, format
    allow_unbounded: bool	= True,
    passphrase: Optional[Union[bytes,str]] = None,	# If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
    format: Optional[str]	= None,			# If the default format for every cryptopath isn't desired
    edit: Optional[str]		= None,
    hardened_defaults: bool	= False,
) -> Sequence[Sequence[Account]]:
    """Generate the desired cryptocurrency account(s) at each crypto's given path(s).  This is useful
    for generating sequences of groups of wallets for multiple cryptocurrencies, eg. for receiving
    multiple cryptocurrencies for each client.  Since each cryptocurrency uses a different BIP-44 path,
    we have to generate different sequences.

    Supports ranges in each path segment, eg.:

        ('ETH', "m/44'/60'/0'/0/-")	-- generates all accounts for ETH
        ('BTC', "m/44'/0'/0'/0/-")	-- generates all accounts for BTC

        [
          [ "m/44'/60'/0'/0/0", "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"],
          [ "m/44'/0'/0'/0/0", "1MAjc529bjmkC1iCXTw2XMHL2zof5StqdQ"]
        ],
        [
          [ "m/44'/60'/0'/0/1", "0x8D342083549C635C0494d3c77567860ee7456963"],
          [ "m/44'/0'/0'/0/1", "1BGwDuVPJeXDG9upaHvVPds5MXwkTjZoav"]
        ],
        ...

    """
    yield from zip( *[
        accounts(
            master_secret	= master_secret,
            crypto		= cry,
            paths		= pth,
            format		= fmt,
            allow_unbounded	= allow_unbounded,
            passphrase		= passphrase,
            using_bip39		= using_bip39,
        )
        for cry,pth,fmt in cryptopaths_parser(
            cryptopaths,
            edit		= edit,
            hardened_defaults	= hardened_defaults,
            format		= format,
        )
    ])


def address(
    master_secret: Union[str,bytes],
    crypto: str			= None,
    path: str			= None,
    format: Optional[str]	= None,
    passphrase: Optional[Union[bytes,str]] = None,  # If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
):
    """Return the specified cryptocurrency HD account address at path."""
    return account(
        master_secret,
        path		= path,
        crypto		= crypto,
        format		= format,
        passphrase	= passphrase,
        using_bip39	= using_bip39,
    ).address


def addresses(
    master_secret: Union[str,bytes],
    crypto: str	 		= None,  # default 'ETH'
    paths: str			= None,  # default: The crypto's path_default; supports ranges
    format: Optional[str]	= None,
    allow_unbounded: bool	= True,
    passphrase: Optional[Union[bytes,str]] = None,  # If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
):
    """Generate a sequence of cryptocurrency account (path, address, ...)  for all designated
    cryptocurrencies.  Usually a single (<path>, <address>) tuple is desired (different
    cryptocurrencies typically have their own unique path derivations.

    """
    for acct in accounts(
            master_secret,
            crypto	= crypto,
            paths	= paths,
            format	= format,
            allow_unbounded = allow_unbounded,
            passphrase	= passphrase,
            using_bip39	= using_bip39,
    ):
        yield (acct.crypto, acct.path, acct.address)


def addressgroups(
    master_secret: Union[str,bytes],
    cryptopaths: Optional[Sequence[Union[str,Tuple[str,str],Tuple[str,str,str]]]] = None,  # default: ETH, BTC at default path, format
    allow_unbounded: bool	= True,
    passphrase: Optional[Union[bytes,str]] = None,  # If mnemonic(s) provided, then passphrase/using_bip39 optional
    using_bip39: bool		= False,
    format: Optional[str]	= None,
    edit: Optional[str]		= None,
    hardened_defaults: bool	= False,
) -> Sequence[str]:
    """Yields account (<crypto>, <path>, <address>) records for the desired cryptocurrencies at paths.

    """
    yield from zip( *[
        addresses(
            master_secret	= master_secret,
            crypto		= cry,
            paths		= pth,
            format		= fmt,
            allow_unbounded	= allow_unbounded,
            passphrase		= passphrase,
            using_bip39		= using_bip39,
        )
        for cry,pth,fmt in cryptopaths_parser(
            cryptopaths,
            edit		= edit,
            hardened_defaults	= hardened_defaults,
            format		= format,
        )
    ])




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__          import annotations

import click
import json
import logging
import string

from ..			import addresses as slip39_addresses
from ..util		import commas, log_cfg, log_level, input_secure
from ..defaults		import BITS

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

"""
Provide basic CLI access to the slip39 API.

Output generally defaults to JSON.  Use -v for more details, and --no-json to emit standard text output instead.
"""

log				= logging.getLogger( __package__ )


@click.group()
@click.option('-v', '--verbose', count=True)
@click.option('-q', '--quiet', count=True)
@click.option( '--json/--no-json', default=True, help="Output JSON (the default)")
def cli( verbose, quiet, json ):
    cli.verbosity		= verbose - quiet
    log_cfg['level']		= log_level( cli.verbosity )
    logging.basicConfig( **log_cfg )
    if verbose or quiet:
        logging.getLogger().setLevel( log_cfg['level'] )
    cli.json			= json
cli.verbosity			= 0  # noqa: E305
cli.json			= False


@click.command()
@click.option( "--crypto", help="The cryptocurrency address to generate (default: BTC)" )
@click.option( "--paths", help="The HD wallet derivation path (default: the standard path for the cryptocurrency; if xpub, omits leading hardened segments by default)" )
@click.option( "--secret", required=True, help="A hex seed or '{x,y,z}{pub,prv}...' x-public/private key to derive HD wallet addresses from; '-' reads it from stdin" )
@click.option( "--format", help="legacy, segwit, bech32 (default: standard for cryptocurrency or '{x,y,z}{pub/prv}...' key)" )
@click.option( '--unbounded/--no-unbounded', default=False, help="Allow unbounded sequences of addresses")
def addresses( crypto, paths, secret, format, unbounded ):
    if secret == '-':
        secret			= input_secure( 'Master secret hex: ', secret=True )
    elif secret and ( secret.lower().startswith( '0x' )
                      or all( c in string.hexdigits for c in secret )):
        log.warning( "It is recommended to not use '-s|--secret <hex>'; specify '-' to read from input" )
    if secret and secret.lower().startswith('0x'):
        secret			= secret[2:]
    if not secret:
        log.error( f"Provide a random {commas( BITS, final='or' )}-bit Seed via --secret" )
    if cli.json:
        click.echo( "[" )
    for i,(cry,pth,adr) in enumerate( slip39_addresses(
            master_secret	= secret,
            crypto		= crypto,
            paths		= paths,
            format		= format,
            allow_unbounded	= unbounded
    )):
        if cli.json:
            if i:
                click.echo( "," )
            if cli.verbosity > 0:
                click.echo( f"    [{json.dumps( cry )+',':6} {json.dumps( pth )+',':21} {json.dumps( adr )}]", nl=False )
            else:
                click.echo( f"    {json.dumps( adr )}", nl=False )
        else:
            if cli.verbosity > 0:
                click.echo( f"{cry:5} {pth:20} {adr}" )
            else:
                click.echo( f"{adr}" )
    if cli.json:
        click.echo( "\n]" )


cli.add_command( addresses )



import os
import sys
import glob
import fnmatch

# 
# All platforms
# 
HERE				= os.path.dirname( os.path.abspath( __file__ ))

install_requires		= open( os.path.join( HERE, "requirements.txt" )).readlines()
tests_require			= open( os.path.join( HERE, "requirements-tests.txt" )).readlines()
options_require			= [
        'gui',		# slip39[gui]:		Support PySimpleGUI/tkinter Graphical UI App
        'serial',	# slip39[serial]:	Support serial I/O of generated wallet data
        'wallet',	# slip39[wallet]:	Paper Wallet and BIP-38/Ethereum wallet encryption
      # 'invoice',	# slip39[invoice]:	Generation of invoices, and associated Smart Contracts
]
extras_require			= {
    option: list(
        # Remove whitespace, elide blank lines and comments
        ''.join( r.split() )
        for r in open( os.path.join( HERE, f"requirements-{option}.txt" )).readlines()
        if r.strip() and not r.strip().startswith( '#' )
    )
    for option in options_require
}
# Make python-slip39[all] install all extra (non-tests) requirements, excluding duplicates
extras_require['all']		= list( set( sum( extras_require.values(), [] )))

Executable			= None
if sys.platform == 'win32':
    # We use cx_Freeze for executable/installer packaging on Windows, only, for now.
    from cx_Freeze import setup, Executable
    print( f"sys.platform {sys.platform}: Using cx_Freeze.setup" )
else:
    from setuptools import setup
    print( f"sys.platform {sys.platform}: Using setuptools.setup" )

# 
# For various platforms, extra setup options are required.  Collect them here
# 
extra_options			= {}

# 
# cx_Freeze options -- Windows .exe/.msi support
#     cx-freeze.readthedocs.io/en/latest/setup_script.html
# 
mainscript			= "SLIP-39.py"
copyright			= "Copyright (c) 2022 Perry Kundert",

if sys.platform == 'win32':
    company_name		= "pjkundert"
    product_name		= "python-slip39"
    icon_win			= "images/SLIP-39.ico"

    shortcut			= (
        "DesktopShortcut",		# Shortcut
        "DesktopFolder",		# Directory_
        "SLIP-39",			# Name
        "TARGETDIR",			# Component_
        "[TARGETDIR]SLIP-39.exe",	# Target
        None,				# Arguments
        None,				# Description
        None,				# Hotkey
        None,				# Icon
        None,				# IconIndex
        None,				# ShowCmd
        "TARGETDIR",			# WkDir
       )

    msi_data			= dict(
        Shortcut	= [
        	shortcut,
        ],
        #Icon		= [
        #    icon_win,
        #]
       )

    bdist_msi_options		= dict(
        add_to_path	= True,
        data		= msi_data,
        initial_target_dir = rf"[ProgramFilesFolder]\{company_name}\{product_name}",
    )

    build_exe_options		= dict(
        packages	= [
            'eth_account', 'Crypto', 'hdwallet', 'shamir_mnemonic', 'cytoolz', 'eth_hash',
        ],
        excludes	= [],
        include_msvcr	= True,
    )

    executables			= [
        Executable(
            mainscript,
            copyright	= copyright,
            base	= "Win32GUI",
            icon	= icon_win,
        ),
    ]

    extra_options		= dict(
        executables	= executables,
        options		= dict(
            bdist_msi	= bdist_msi_options,
            build_exe	= build_exe_options,
        )
    )


'''
if sys.platform == 'darwin':
    # For py2{app,exe} App Generation.  TODO: Does not work; use PyInstaller instead
    extra_options		= dict(
        setup_requires	= [ 'py2app' ],
        app		= [ mainscript ],
        # Cross-platform applications generally expect sys.argv to
        # be used for opening files.
        # Don't use this with GUI toolkits, the argv
        # emulator causes problems and toolkits generally have
        # hooks for responding to file-open events.
        options		= dict(
            py2app	= dict(
                argv_emulation	= True,
                iconfile	= 'images/SLIP39.icns',
                includes	= 'tkinter',
            ),
        ),
    )
elif sys.platform == 'win32':
    extra_options		= dict(
        setup_requires	= [ 'py2exe' ],
        app		= [ mainscript ],
    )
else:
    extra_options		= dict(
        # Normally unix-like platforms will use "setup.py install"
        # and install the main script as such
        scripts		= [ mainscript ],
    )
'''

# Must work if setup.py is run in the source distribution context, or from
# within the packaged distribution directory.
__version__			= None
try:
    exec( open( 'slip39/version.py', 'r' ).read() )
except FileNotFoundError:
    exec( open( 'version.py', 'r' ).read() )

console_scripts			= [
    'slip39		= slip39.main:main',
    'slip39-recovery	= slip39.recovery.main:main',
    'slip39-generator	= slip39.generator.main:main',
    'slip39-gui		= slip39.gui.main:main',
    'slip39-cli		= slip39.cli:cli',
]

entry_points			= {
    'console_scripts': 		console_scripts,
}

package_dir			= {
    "slip39":			"./slip39",
    "slip39.cli":		"./slip39/cli",
    "slip39.invoice":		"./slip39/invoice",
    "slip39.layout":		"./slip39/layout",
    "slip39.layout.font":	"./slip39/layout/font",
    "slip39.recovery":		"./slip39/recovery",
    "slip39.generator":		"./slip39/generator",
    "slip39.gui":		"./slip39/gui",
    # Until tabulate fixes empty cell type deduction, separators
    "slip39.tabulate":		"./slip39/tabulate",
}

package_data			= {
    'slip39': [
        'layout/*.png',
        'layout/*.txt',
        'layout/font/*.ttf',
        'gui/*.txt',
    ],
}

long_description_content_type	= 'text/markdown'
long_description		= """\
Creating Ethereum, Bitcoin and other accounts is complex and fraught
with potential for loss of funds.

A BIP-39 seed recovery phrase helps, but a *single* lapse in security
dooms the account (and all derived accounts, in fact).  If someone finds
your recovery phrase (or you lose it), the accounts derived from that
seed are /gone/.

The SLIP-39 standard allows you to split the seed between 1, 2, or more
groups of several mnemonic recovery phrases.  This is better, but
creating such accounts is difficult; presently, only the Trezor supports
these, and they can only be created "manually".  Writing down 5 or more
sets of 20 words is difficult, error-prone and time consuming.

# Hardware Wallet "Seed" Configuration

>  Your keys, your Bitcoin.  Not your keys, not your Bitcoin.
>  
>  ---Andreas Antonopoulos

The [python-slip39][1] project (and the [slip39-app][2]) exists to assist in the safe creation,
backup and documentation of Hierarchical Deterministic (HD) Wallet seeds and derived accounts,
with various SLIP-39 sharing parameters.  It generates the new random wallet seed, and generates
the expected standard Ethereum account(s) (at *m/44'/60'/0'/0/0* by default) and
Bitcoin accounts (at Bech32 derivation path *m/84'/0'/0'/0/0* by default), with wallet address and
QR code (compatible with Trezor and Ledger derivations).  It produces the required SLIP-39
phrases, and outputs a single PDF containing all the required printable cards to document the seed
(and the specified derived accounts).

On an secure (ideally air-gapped) computer, new seeds can /safely/ be generated (*without*
trusting this program) and the PDF saved to a USB drive for printing (or directly printed without
the file being saved to disk.).  Presently, =slip39= can output example ETH, BTC, LTC, DOGE, BSC,
and XRP addresses derived from the seed, to /illustrate/ what accounts are associated with the
backed-up seed.  Recovery of the seed to a [trezor-model-t][3] is simple, by entering the mnemonics
right on the device.

We also support the backup of existing insecure and unreliable 12- or 24-word BIP-39 Mnemonic
Phrases as SLIP-39 Mnemonic cards, for existing BIP-39 hardware wallets like the [ledger-nano][4],
etc.!  Recover from your existing BIP-39 Seed Phrase Mnemonic, select "Using BIP-39" (and enter
your BIP-39 passphrase), and generate a set of SLIP-39 Mnemonic cards.  Later, use the SLIP-39 App
to recover from your SLIP-39 Mnemonic cards, click "Using BIP-39" to get your BIP-39 Mnemonic
back, and use it (and your passphrase) to recover your accounts to your Ledger (or other) hardware
wallet.

Output of BIP-38 or JSON encrypted Paper Wallets is also supported, for import into standard
software cryptocurrency wallets.

[1]: <https://github.com/pjkundert/python-slip39.git> "python-slip39"

[2]: <https://slip39.com/app> "SLIP-39 App"

[3]: <https://trezor.go2cloud.org/SH1Y> "Trezor Model T"

[4]: <https://shop.ledger.com/pages/ledger-nano-x?r=2cd1cb6ae51f> "Ledger Nano"

## The SLIP-39 App for Generation and Backup

Download the SLIP-39 App from [slip39-download][5] for Windows (.msi) or macOS (.dmg).

The Graphical User Interface (GUI) also supports Linux:

    $ python3 -m pip install slip39[gui]
    $ slip39-gui

![SLIP-39 App GUI][slip39-gui]

[5]: <https://github.com/pjkundert/python-slip39/releases/latest> "SLIP-39 App Download"
[slip39-gui]: https://github.com/pjkundert/python-slip39/raw/master/images/slip39.png "SLIP-39 App GUI"

## Generating SLIP-39 on the Command Line

    $ python3 -m slip39 -v Personal      # or run: slip39 -v Personal
    2022-01-26 13:55:30 slip39           First(1/1): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 cricket   15 unhappy
    2022-01-26 13:55:30 slip39                2 acid       9 mental    16 ocean
    2022-01-26 13:55:30 slip39                3 acrobat   10 veteran   17 mayor
    2022-01-26 13:55:30 slip39                4 romp      11 phantom   18 promise
    2022-01-26 13:55:30 slip39                5 anxiety   12 grownup   19 wrote
    2022-01-26 13:55:30 slip39                6 laser     13 skunk     20 romp
    2022-01-26 13:55:30 slip39                7 cricket   14 anatomy
    2022-01-26 13:55:30 slip39           Second(1/1): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 belong    15 spirit
    2022-01-26 13:55:30 slip39                2 acid       9 survive   16 royal
    2022-01-26 13:55:30 slip39                3 beard     10 home      17 often
    2022-01-26 13:55:30 slip39                4 romp      11 herd      18 silver
    2022-01-26 13:55:30 slip39                5 again     12 mountain  19 grocery
    2022-01-26 13:55:30 slip39                6 orbit     13 august    20 antenna
    2022-01-26 13:55:30 slip39                7 very      14 evening
    2022-01-26 13:55:30 slip39           Fam(2/4): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 rainbow   15 husky
    2022-01-26 13:55:30 slip39                2 acid       9 swing     16 crowd
    2022-01-26 13:55:30 slip39                3 ceramic   10 credit    17 learn
    2022-01-26 13:55:30 slip39                4 roster    11 piece     18 priority
    2022-01-26 13:55:30 slip39                5 already   12 puny      19 hand
    2022-01-26 13:55:30 slip39                6 quiet     13 senior    20 watch
    2022-01-26 13:55:30 slip39                7 erode     14 listen
    2022-01-26 13:55:30 slip39           2nd  1 sister     8 holy      15 revenue
    2022-01-26 13:55:30 slip39                2 acid       9 execute   16 junction
    2022-01-26 13:55:30 slip39                3 ceramic   10 lift      17 elite
    2022-01-26 13:55:30 slip39                4 scared    11 spark     18 flexible
    2022-01-26 13:55:30 slip39                5 domestic  12 yoga      19 inform
    2022-01-26 13:55:30 slip39                6 exact     13 medical   20 predator
    2022-01-26 13:55:30 slip39                7 finger    14 grief
    ...
    2022-01-26 13:55:30 slip39           ETH    m/44'/60'/0'/0/0    : 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1
    2022-01-26 13:55:30 slip39           BTC    m/84'/0'/0'/0/0     : bc1q6u7qk0tepkxdm8wkhpqzwwy0w8zfls9yvghaxq
    ...
    2022-01-26 13:55:30 slip39           Wrote SLIP39-encoded wallet for 'Personal' to: Personal-2022-01-26+13.55.30-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.pdf

## Recover from SLIP-39 Mnemonics

Later, if you need to recover the Ethereum wallet, keep entering SLIP-39 mnemonics until the secret
is recovered (invalid/duplicate mnemonics will be ignored):

    $ python3 -m slip39.recovery -v      # or run: slip39-recovery -v
    Enter 1st SLIP-39 mnemonic: sister acid acrobat romp anxiety laser cricket cricket mental veteran phantom grownup skunk anatomy unhappy ocean mayor promise wrote romp
    2021-12-29 13:24:25 slip39.recovery  Could not recover SLIP-39 master secret with 1 supplied mnemonics: Insufficient number of mnemonic groups. The required number of groups is 2.
    Enter 2nd SLIP-39 mnemonic: a bc
    2021-12-29 13:24:53 slip39.recovery  Could not recover SLIP-39 master secret with 2 supplied mnemonics: Invalid mnemonic word 'a'.
    Enter 3rd SLIP-39 mnemonic: sister acid ceramic roster already quiet erode rainbow swing credit piece puny senior listen husky crowd learn priority hand watch
    2021-12-29 13:24:58 slip39.recovery  Could not recover SLIP-39 master secret with 3 supplied mnemonics: Invalid mnemonic word 'a'.
    Enter 4th SLIP-39 mnemonic: sister acid ceramic scared domestic exact finger holy execute lift spark yoga medical grief revenue junction elite flexible inform predator
    2021-12-29 13:25:14 slip39.recovery  Recovered SLIP-39 secret with 3 (1st, 3rd, 4th) of 4 supplied mnemonics
    2021-12-29 13:25:14 slip39.recovery  Recovered SLIP-39 secret; To re-generate, send it to: python3 -m slip39 --secret -
    32448aabb50cb6b022fdf17d960720df

Finally, regenerate the Ethereum wallet, perhaps including an encrypted JSON wallet file for import
into a software wallet; note that the same Ethereum wallet address 0x8FBC...3bf1 is recovered:

    $ python3 -m slip39 --secret 32448aabb50cb6b022fdf17d960720df --json -
    2022-01-26 14:06:14 slip39           It is recommended to not use '-s|--secret <hex>'; specify '-' to read from input
    2022-01-26 14:06:14 slip39           ETH    m/44'/60'/0'/0/0    : 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1
    2022-01-26 14:06:14 slip39           BTC    m/84'/0'/0'/0/0     : bc1q6u7qk0tepkxdm8wkhpqzwwy0w8zfls9yvghaxq
    JSON key file password:
    2022-01-26 14:06:21 slip39           Wrote JSON SLIP39's encrypted ETH wallet 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1 \
                                                derived at m/44'/60'/0'/0/0 to: SLIP39-2022-01-26+14.06.14-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.json
    2022-01-26 14:06:21 slip39           Wrote SLIP39-encoded wallet for '' to: SLIP39-2022-01-26+14.06.14-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.pdf

The whole toolchain is suitable for pipelining:

    $ python3 -m slip39 --text --no-card -q \\
        | sort -r \\
        | python3 -m slip39.recovery \\
        | python3 -m slip39 --secret - --no-card -q
    2021-12-28 10:55:17 slip39           ETH m/44'/60'/0'/0/0    : 0x68dD9B59D5dF605f4e9612E8b427Ab31187E2C54
    2021-12-28 10:55:18 slip39.recovery  Recovered SLIP-39 secret with 4 (1st, 2nd, 7th, 8th) of 8 supplied mnemonics
    2021-12-28 10:55:18 slip39           ETH m/44'/60'/0'/0/0    : 0x68dD9B59D5dF605f4e9612E8b427Ab31187E2C54

## SLIP-39 Mnemonic Cards PDF

Here's an example of PDF containing the SLIP-39 recovery mnemonic cards produced:

![SLIP-39 Mnemonic Cards PDF][slip39-cards]

If desired, you can output encrypted "Paper Wallets", compatible for import into standard software
cryptocurrency wallets (BIP-38 for Bitcoin, etc., and standard JSON wallets for Ethereum):

![Encrypted Paper Wallets PDF][slip39-wallets]

[slip39-cards]: https://github.com/pjkundert/python-slip39/raw/master/images/slip39-cards.png "SLIP-39 Mnemonic Cards PDF"
[slip39-wallets]: https://github.com/pjkundert/python-slip39/raw/master/images/slip39-wallets.png "Encrypted Paper Wallets PDF"
"""

classifiers			= [
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "License :: Other/Proprietary License",
    "Programming Language :: Python :: 3",
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Financial and Insurance Industry",
    "Environment :: Console",
    "Topic :: Security :: Cryptography",
    "Topic :: Office/Business :: Financial",
]
project_urls			= {
    "Bug Tracker": "https://github.com/pjkundert/python-slip39/issues",
}

setup(
    name			= "slip39",
    version			= __version__,
    install_requires		= install_requires,
    tests_require		= tests_require,
    extras_require		= extras_require,
    packages			= package_dir.keys(),
    package_dir			= package_dir,
    package_data		= package_data,
    include_package_data	= True,
    zip_safe			= True,
    entry_points		= entry_points,
    author			= "Perry Kundert",
    author_email		= "perry@dominionrnd.com",
    project_urls		= project_urls,
    description			= "Standards-compliant SLIP-39 cryptocurrency seed generation and recovery, compatible with Trezor hardware wallets",
    long_description		= long_description,
    long_description_content_type = long_description_content_type,
    license			= "Dual License; GPLv3 and Proprietary",
    keywords			= "Ethereum Bitcoin Dogecoin Litecoin cryptocurrency SLIP-39 BIP-39 seed recovery PDF BIP-38 paper wallet",
    url				= "https://github.com/pjkundert/python-slip39",
    classifiers			= classifiers,
    python_requires		= ">=3.9",
    **extra_options
)



import logging
import os
import re
import sys

from io			import StringIO
from pathlib		import Path
from subprocess		import Popen, PIPE
from email		import message_from_string

import dkim

from aiosmtpd.controller import Controller

from .communications	import dkim_message, send_message, matchaddr, AutoResponder
from .defaults		import SMTP_TO, SMTP_FROM

log				= logging.getLogger( __package__ )

# Disable printing of details unless something goes wrong...
print_NOOP			= lambda *args, **kwds: None
print				= print_NOOP

# If we find a DKIM key, lets use it.  Otherwise, we'll just use the pre-defined pre-signed email.Message
dkim_keys			= list( Path( __file__ ).resolve().parent.parent.glob( 'licensing.dominionrnd.com.*.key' ))
dkim_key			= None
dkim_msg			= None
if dkim_keys:
    # Choose the latest available, assuming YYYYMMDD selector
    dkim_key			= str( sorted( dkim_keys )[-1] )
    dkim_selector		= re.match( r'.*\.(\d+)\.key', dkim_key ).group( 1 )
else:
    # No local key (the usual case, unless you're Dominion R&D!)
    dkim_selector		= '20221230'
    dkim_msg			= message_from_string( """\
Content-Type: multipart/alternative; boundary================1903566236404015660==
MIME-Version: 1.0
From: no-reply@licensing.dominionrnd.com
To: licensing@dominionrnd.com
Reply-To: perry@kundert.ca
Subject: Hello, world!
DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/simple; d=licensing.dominionrnd.com; i=@licensing.dominionrnd.com; q=dns/txt; s=20221230; t=1673405994; h=from : to;\
 bh=RkF6KP4Q94MVDBEv7pluaWdzw0z0GNQxK72rU02XNcE=; b=Tao30CJGcqyX86f37pSrSFLSDvA8VkzQW0jiMf+aFg5D99LsUYmUZxSgnDhW2ZEzjwu6bzjkEEyvSEv8LxfDUW+AZZG3enbq/mnnUZw3PXp4l\
MaZGN9whvTIUy4/QUlMGKuf+7Vzi+8eKKjh4CWKN/UEyX6YoU7V5eyjTTA7q1jIjEl8jiM4LXYEFQ9LaKUmqqmRh2OkxBVf1QG+fEYTYUed+oS05m/d1SyVLjxv8ldeXT/mGgm1CrGk1qfRTzfcksX4qNAluTfJTa\
kDpHNPw0RX0QzkuWvgWG5GngV65yg6fL87wQOVqV4O7OhK6eTkzWqzNyerJd4i6B7ZCoYEUg==

--===============1903566236404015660==
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

Testing 123
--===============1903566236404015660==
Content-Type: text/html; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

<em>Testing 123</em>
--===============1903566236404015660==--

""" )


def test_communications_matchaddr():
    assert matchaddr( "abc+def@xyz", mailbox="abc", domain="xyz" ) == ("abc", "def", "xyz")
    assert matchaddr( "abc+def@xyz",                domain="xYz" ) == ("abc", "def", "xyz")
    assert matchaddr( "abc+def@xyz", mailbox="Abc"               ) == ("abc", "def", "xyz")
    assert matchaddr( "abc+def@xyz",                             ) == ("abc", "def", "xyz")
    assert matchaddr( "abc+def@xyz", "a*c","*f","x?z"            ) == ("abc", "def", "xyz")
    assert matchaddr( "abc+def@xyz", "b*c","*f","x?z"            ) is None
    assert matchaddr( "abc+def@xyz", "a*c","*f","x?"             ) is None
    assert matchaddr( "abc+def@xyz",                             )[2] == "xyz"
    assert matchaddr( "abc+def@xyz", mailbox="xxx"               ) is None


def test_communications_dkim():
    log.info( f"Using DKIM: {dkim_selector}: {dkim_key}" )
    if dkim_key:
        msg			= dkim_message(
            sender_email	= SMTP_FROM,			# Message From: specifies claimed sender
            to_email		= SMTP_TO,			# See https://dkimvalidator.com to test!
            reply_to_email	= "perry@kundert.ca",
            subject		= "Hello, world!",
            message_text	= "Testing 123",
            message_html	= "<em>Testing 123</em>",
            dkim_private_key_path = dkim_key,
            dkim_selector	= dkim_selector,
            headers		= ['From', 'To'],
        )
    else:
        msg			= dkim_msg

    log.info( f"DKIM {'signed' if dkim_key else 'canned'} Message:\n{msg}" )

    sig			= msg['DKIM-Signature']
    sig_kvs		= sig.split( ';' )
    sig_k_v		= dict(
        (k.strip(), v.strip())
        for k,v in ( kv.split( '=', 1 ) for kv in sig_kvs )
    )
    assert sig_k_v['v'] == '1'
    assert sig_k_v['a'] in ( 'rsa-sha256', 'ed25519-sha256' )

    assert dkim.verify( msg.as_bytes() )

    # Send via port 587 w/ TLS.  Use the appropriate relay servers.  The cloudflare MX servers only
    # respond on port 25; TLS port 587 don't see to respond Since port 25 is usually blocked for
    # most retail internet connections, we'll use our own TLS-capable relays.  So: for normal,
    # retail ISP-connected hosts, sending SMTP email via port 25 is usually impossible.  Via port
    # 587 (TLS) or 465 (SSL), it might be possible.
    #
    # Altering the MAIL FROM: (and hence the email's Return-Path), you can alter the behaviour of
    # auto-responders.  However, SPF analysis looks at the Return-Path to determine which domain's
    # TXT records to examine!  Therefore, the act of changing the MAIL FROM: address to the intended
    # recipient of the auto-response, makes it certain that any SPF record on the intended
    # recipient's domain will mark this email with an Received-SPF: fail!
    #
    # There doesn't appear to be a strictly "legal" method in SMTP to designate the intended
    # recipient of the response.  I guess, to avoid using "bounces" to send SPAM email.  Therefore,
    # the auto-responder must be programmed to use the Reply-To address -- something that eg. Gmail
    # cannot be programmed to do.
    try:
        send_message(
            msg,
            #from_addr	= SMTP_FROM,		# Envelope MAIL FROM: specifies actual sender
            #to_addrs	= [ SMTP_TO ],		# Will be the same as message To: (should default)
            #relay		= ['mail2.kundert.ca'],  # 'localhost',   # use eg. ssh -fNL 0.0.0.0:25:linda.mx.cloudflare.net:25 root@your.VPS.com
            #port		= 25,  # 465 --> SSL, 587 --> TLS (default),
            #usessl = False, starttls = False, verifycert = False,  # to mail.kundert.ca; no TLS
            #usessl = False, starttls = True, verifycert = False,  # default
        )
    except Exception as exc:
        # This may fail (eg. if you have no access to networking), so we don't check.
        log.warning( f"(Expected, if no networking): Failed to send DKIM-validated email to {SMTP_TO}: {exc}" )
        pass


def test_communications_autoresponder( monkeypatch ):
    """The Postfix-compatible auto-responder takes an email.Message from stdin, and auto-forwards it
    (via a relay; normally the same Postfix installation that it is running within).

    Let's shuttle a simple message through the AutoResponder, and fire up an SMTP daemon to receive
    the auto-forwarded message(s).

    """
    if dkim_key:
        msg			= dkim_message(
            sender_email	= SMTP_FROM,			# Message From: specifies claimed sender
            to_email		= SMTP_TO,			# See https://dkimvalidator.com to test!
            reply_to_email	= "perry@kundert.ca",
            subject		= "Hello, world!",
            message_text	= "Testing 123",
            message_html	= "<em>Testing 123</em>",
            dkim_private_key_path = dkim_key,
            dkim_selector	= dkim_selector,
            headers		= ['From', 'To'],
        )
    else:
        msg			= dkim_msg

    envelopes		= []

    class PrintingHandler:
        async def handle_RCPT(self, server, session, envelope, address, rcpt_options):
            if matchaddr( address )[2] not in ('dominionrnd.com', 'kundert.ca'):
                return f'550 not relaying to {address}'
            envelope.rcpt_tos.append(address)
            return '250 OK'

        async def handle_DATA(self, server, session, envelope):
            print('Message from %s' % envelope.mail_from)
            print('Message for %s' % envelope.rcpt_tos)
            print('Message data:\n')
            for ln in envelope.content.decode('utf8', errors='replace').splitlines():
                print(f'> {ln}'.strip())
            print()
            print('End of message')
            envelopes.append( envelope )
            return '250 Message accepted for delivery'

    handler		= PrintingHandler()
    # dynamic port allocation w/ port=0 not scheduled 'til ~v1.5
    controller		= Controller( handler, hostname='localhost', port=11111 )
    controller.start()

    # Send the email.Message directly our SMTP daemon, w/ RCTP TO: licensing@dominionrnd.com (taken
    # from the To: header)
    send_message(
        msg,
        relay		= controller.hostname,
        port		= controller.port,
        starttls	= False,
        usessl		= False
    )
    assert len( envelopes ) == 1
    assert envelopes[-1].rcpt_tos == [ 'licensing@dominionrnd.com' ]

    # Now, run the auto-responder on messages to licensing@dominionrnd.com, not reinjecting the
    # message into the mail system, and sending the auto-response to the SMTP daemon.  We should see
    # it at the Reply-To: perry@kundert.ca header address.
    from_addr			= msg['From'] or msg['Sender']
    to_addrs			= []
    if 'To' in msg:
        to_addrs.append( msg['To'] )
    for cc in ('Cc', 'Bcc'):
        if cc in msg:
            to_addrs	       += map( str.strip, msg[cc].split( ',' ) )

    monkeypatch.setattr( 'sys.stdin', StringIO( msg.as_string() ))
    ar				= AutoResponder(
        address		= SMTP_FROM,
        server		= controller.hostname,
        port		= controller.port,
        reinject	= False,
    )
    status			= ar(
        from_addr, *to_addrs
    )
    assert status == 0
    assert len( envelopes ) == 2
    assert envelopes[-1].rcpt_tos == [ 'perry@kundert.ca' ]

    # Now, try the CLI version.  Must use the current python interpreter, and this local instance of python-slip39
    here			= Path( __file__ ).resolve().parent

    for execute in [
        [
            sys.executable, "-m", "slip39.communications",
        ]

    ]:
        command			= list( map( str, execute + [
            '-vv', '--no-json',
            'autoresponder',
            '--server',		controller.hostname,
            '--port',		controller.port,
            #'--no-reinject',
            '--reinject',	"echo",
            '*@*licensing.dominionrnd.com',  # allow any sender mailbox, any ...licensing subdomain of dominionrnd.com
            from_addr,
            *to_addrs
        ] ))
        PYTHONPATH		= f"{here.parent}"
        log.info( f"Running filter w/ PYTHONPATH={PYTHONPATH}: {' . '.join( command )}" )
        with Popen(
                command,
                stdin	= PIPE,
                stdout	= PIPE,
                stderr	= PIPE,
                env	= dict(
                    os.environ,
                    PYTHONPATH	= PYTHONPATH,
                )) as process:
            out, err		= process.communicate( msg.as_bytes() )
            log.info( f"Filter stdout: {out.decode( 'UTF-8' ) if out else out}, stderr: {err.decode( 'UTF-8' ) if err else err}" )
            assert process.returncode == 0
    assert len( envelopes ) == 3
    assert envelopes[-1].rcpt_tos == [ 'perry@kundert.ca' ]



#! /usr/bin/env python3
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
import json
import logging
import re
import smtplib
import ssl
import sys

from fnmatch		import fnmatch
from subprocess		import Popen, PIPE

import click
import dkim

from crypto_licensing.licensing import doh
from email		import utils, message_from_file, message_from_string
from email.mime		import multipart, text

#from tabulate		import tabulate
from .tabulate		import tabulate

from .util		import is_listlike, commas, uniq, log_cfg, log_level


__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2023 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( 'email' )


def mx_records( domain, timeout=None ):
    """Query and yield the MX records for 'domain', via DoH, sorted by their priority."""
    kwds			= {}
    if timeout:
        kwds.update( timeout=timeout )
    for _,mx in sorted(
        mx['data'].split()
        for mx in doh.query(
            domain	= domain,
            record	= doh.DNSRecord.MX,
            **kwds
        )
    ):
        yield mx


def matchglob( string, pattern ):
    """Returns the 'string' matching the Glob 'pattern', nor None."""
    if set( '*?[' ) & set( pattern ):
        if fnmatch( string, pattern ):
            return string
    elif pattern == string:
        return string


def matchaddr( address, mailbox=None, extension=None, domain=None ):
    """Parse and email address (eg. from "Real Name <mailbox+ext@domain.com>"), and return its
    (<mailbox>,<extension>,<domain>), or None if no address parsed/matched.

    The supplied email address begins with "<mailbox>", optionally followed by a "+<extension>", and
    then ends with "@<domain>".  If so, return the re.match's 3 groups.  Note: a group may be None
    if not present, ie. "someone@domain.com" returns ("someone", None, "domain.com").  If 'mailbox',
    'extension' or 'domain' are falsey, any will be allowed.

    Allows Glob-style patterns in the supplied 'mailbox', 'extension' or 'domain', eg:

        domain="dominion*.c*"

    would match email addresses to both "...@dominionrnd.com" and "...@dominionwheat.ca"

    Does not properly respect email addresses with quoting, eg. 'abc"123@456"@domain.com' because,
    quite frankly, I don't want to and that's just "Little Bobby Tables" (https://xkcd.com/327/)
    level asking for trouble...

    Simple <mailbox>[+<extension>]@<domain>, please.

    """
    m			= re.match(
        r"(^[^@+]+)(?:\+([^@]+))?@(.*)",
        utils.parseaddr( address )[1],
    )
    if ( not mailbox or ( m.group( 1 ) and matchglob( m.group( 1 ).lower(), mailbox.lower() ))):
        if ( not extension or ( m.group( 2 ) and matchglob( m.group( 2 ).lower(), extension.lower() ))):
            if ( not domain or ( m.group( 3 ) and matchglob( m.group( 3 ).lower(), domain.lower() ))):
                return m.groups()


def dkim_message(
    sender_email,			# The message From:; may be empty (but not for DKIM). For display by the client
    subject,
    message_text,
    message_html	= None,
    to_email		= None,		# The message To: "" ; may be "address, address".  For display by client
    cc_emails		= None,
    bcc_emails		= None,
    reply_to_email	= None,
    dkim_private_key_path="",
    dkim_selector	= "",
    signature_algorithm	= None,
    headers		= None,		# List of headers to DKIM-sign
):
    """

    Create a DKIM-signed email message.


    The `email` library assumes it is working with string objects.
    The `dkim` library assumes it is working with byte objects.
    This function performs the acrobatics to make them both happy.

   See:
       https://github.com/russellballestrini/miscutils/blob/master/miscutils/mail.py
       https://www.emailonacid.com/blog/article/email-deliverability/what_is_dkim_everything_you_need_to_know_about_digital_signatures/

    """

    if isinstance(message_text, bytes):
        # needed for Python 3.
        message_text = message_text.decode()

    if isinstance(message_html, bytes):
        # needed for Python 3.
        message_html = message_html.decode()

    if headers is None:
        headers			= ["From", "To", "Subject"]

    sender_domain		= matchaddr( sender_email )[2]

    msg				= multipart.MIMEMultipart("alternative")
    msg.attach(text.MIMEText(message_text, "plain"))
    if message_html:
        msg.attach(text.MIMEText(message_html, "html"))
    assert sender_email is not None, \
        "DKIM requires the From: address to be signed"
    msg["From"]			= sender_email
    if to_email is not None:
        msg["To"]		= ', '.join( to_email ) if is_listlike( to_email ) else to_email
    if cc_emails is not None:
        msg["Cc"]		= ', '.join( cc_emails ) if is_listlike( cc_emails ) else cc_emails
    if bcc_emails is not None:
        msg["Bcc"]		= ', '.join( bcc_emails ) if is_listlike( bcc_emails ) else bcc_emails
    if reply_to_email:
        # Autoresponders don't generally respect Reply-To (as recommended in RFC-3834)
        # https://www.rfc-editor.org/rfc/rfc3834#section-4.
        msg["Reply-To"]		= reply_to_email
    msg["Subject"]		= subject

    try:
        # Python 3 libraries expect bytes.
        msg_data		= msg.as_bytes()
    except Exception:
        # Python 2 libraries expect strings.
        msg_data		= msg.as_string()

    if dkim_private_key_path and dkim_selector:
        # the dkim library uses regex on byte strings so everything
        # needs to be encoded from strings to bytes.
        with open(dkim_private_key_path) as fh:
            dkim_private_key	= fh.read()
        if signature_algorithm is None:
            # "ed25519-sha256" not well supported, yet.  But, if the key is short, that's probably what it is.
            signature_algorithm	= "rsa-sha256" if len( dkim_private_key ) > 48 else 'ed25519-sha256'

        sig			= dkim.sign(
            message		= msg_data,
            selector		= str(dkim_selector).encode(),
            domain		= sender_domain.encode(),
            privkey		= dkim_private_key.encode(),
            include_headers	= [ h.encode() for h in headers ],
            signature_algorithm	= signature_algorithm.encode(),
        )
        #
        # Unfortunately, the produced:
        #
        #     b'DKIM-Signature: v=1; i=@lic...\r\n s=... b=Fp2...6H\r\n 5//6o...Ag=='
        #                                     ^^^^^                ^^^^^
        #
        # contains a bunch of unnecessary whitespace, especially within the b: and bh: base-64 data.
        # However, this whitespace is ignored by the standard email.Message parser.  Tidy it up.
        #
        pre,sig_dirty		= sig.decode( 'utf-8' ).split( ':', 1 )
        log.info( f"DKIM signed: {sig_dirty.strip()!r}" )
        assert pre.lower() == "dkim-signature"

        # This seemed to corrupt the signature, unexpectedly...  Ah, because we removed internal
        # whitespace from with the '; h=from : to', which is part of the signature.  Only clean b/bh.
        sig_kvs		= sig_dirty.split( ';' )
        sig_k_v		= dict(
            (k.strip(), v.strip())
            for k,v in ( kv.split( '=', 1 ) for kv in sig_kvs )
        )
        for k in ('b', 'bh'):
            if k in sig_k_v:  # eliminates internal whitespace
                sig_k_v[k]	= ''.join( sig_k_v[k].split() )
        sig_clean		= '; '.join( f"{k}={v}" for k,v in sig_k_v.items() )
        log.info( f"DKIM clean:  {sig_clean!r}" )
        # add the dkim signature to the email message headers.
        msg["DKIM-Signature"]	= sig_clean

        return msg


def send_message(
    msg,
    from_addr		= None,		# Envelope MAIL FROM: (use msg['Sender'/'From'] if not specified)
    to_addrs		= None,		# Envelope RCTP TO:   (use msg['To'/'CC'/'BCC'] if not specified)
    relay		= None,		# Eg. "localhost"; None --> lookup MX record
    port		= None,		# Eg. 25 --> raw TCP/IP, 587 --> TLS (default), 465 --> SSL
    starttls		= None,		# Upgrade SMTP connection w/ TLS (default: True iff port == 587)
    usessl		= None,		# Connect using SMTP_SSL (default: True iff port == 465)
    verifycert		= None,		# and verify SSL/TLS certs (not generally supported)
    username		= None,
    password		= None,
):
    """Send a (possibly DKIM-signed) email msg, ideally directly to target domain's SMTP server,
    "RCPT TO" (each) to_addrs recipient, "MAIL FROM" the from_addrs.

    Unless relay(s) specified, we'll look up the MX records for each of the recipient to_addrs.

    """
    # Deduce from address / to addresses from message content, if not supplied.  These are the
    # Envelope MAIL FROM / RCPT TO addresses, which don't necessarily match From or To/CC/BCC.
    # See: smtplib.py's send_message.
    if from_addr is None:
        # Prefer the sender field per RFC 2822:3.6.2.
        from_addr		= msg['Sender'] if 'Sender' in msg else msg['From']
        from_addr		= utils.getaddresses( [from_addr] )[0][1]

    if to_addrs is None:
        addr_fields		= [
            f
            for f in (msg['To'], msg['Bcc'], msg['Cc'])
            if f is not None
        ]
        to_addrs		= [
            a[1]
            for a in utils.getaddresses( addr_fields )
        ]

    # Now that we have a to_addrs, construct a mapping of (mx, ...) --> [addr, ...].  For each
    # to_addrs, lookup its destination's mx records; we'll append all to_addrs w/ the same mx's
    # (sorted by priority).
    relay_addrs			= {}
    if relay is None:
        for to in to_addrs:
            relay_addrs.setdefault( tuple( mx_records( to.split( '@', 1 )[1] )), [] ).append( to )
    else:
        if not is_listlike( relay ):
            relay		= [ relay ]
        relay_addrs[tuple( relay )] = to_addrs
    relay_max			= max( len( r ) for r in relay_addrs.keys() )
    addrs_max			= max( len( a ) for a in relay_addrs.values() )
    log.info( "Relays, and their destination addresses\n" + tabulate(
        [
            list( r ) + ([ None ] * (relay_max - len( r ))) + list( a )
            for r,a in relay_addrs.items()
        ],
        headers= tuple( f"mx {m+1}" for m in range( relay_max ) ) + tuple( f"addr {a+1}" for a in range( addrs_max )),
        tablefmt='orgtbl'
    ))

    # Default port and TLS/SSL if unspecified.
    if port is None:
        port			= 587
    if starttls is None:
        starttls		= True if port == 587 else False
    if usessl is None:
        usessl			= True if port == 465 else False
    if verifycert is None:
        verifycert		= False

    try:
        # Python 3 libraries expect bytes.
        msg_data = msg.as_bytes()
    except Exception:
        # Python 2 libraries expect strings.
        msg_data = msg.as_string()

    smtp_kwds			= dict()
    if usessl:
        smtp_class		= smtplib.SMTP_SSL
        smtp_kwds.update( context = ssl.create_default_context() )
        assert not ( starttls and verifycert ), \
            "starttls w/ verifycert is incompatible with usessl"
    else:
        smtp_class		= smtplib.SMTP

    class smtp_class_logging( smtp_class ):

        def putcmd( self, *args, **kwds ):
            log.info( f"SMTP --> {args!r} {kwds!r}" )
            return super().putcmd( *args, **kwds )

        def getreply( self ):
            try:
                errcode, errmsg = super().getreply()
                log.info( f"SMTP <-- {errcode} {errmsg}" )
                return errcode, errmsg
            except Exception as exc:
                log.info( f"SMTP <x- {exc}" )
                raise

    relayed			= 0
    for ra_i,(mxs, to_addrs) in enumerate( relay_addrs.items() ):
        log.warning( f"Sending {len(to_addrs)} via {mxs}: {to_addrs}" )
        for mx_i,mx in enumerate( mxs ):
            # TODO: react if connecting to relay (localhost postfix) is a socket error.
            log.info( f"Connecting to SMTP server {mx}" )
            try:
                with smtp_class_logging( mx, port, **smtp_kwds ) as mta:
                    if starttls:
                        if verifycert:
                            mta.starttls( context=ssl.create_default_context() )
                        else:
                            mta.starttls()
                    if username and password:
                        mta.auth( mta.auth_plain, user=username, password=password )
                    mta.sendmail( from_addr=from_addr, to_addrs=to_addrs, msg=msg_data )
            except Exception as exc:
                # If this is the last of the relays, and its last MX, and nothing was
                # successfully relayed, then raise the Exception.
                err			= f"Failed to send via SMTP relay {mx}: {exc}"
                log.warning( err )
                if mx_i + 1 == len( mxs ) and ra_i + 1 == len( relay_addrs ) and not relayed:
                    raise RuntimeError( err ) from exc
            else:
                #  Otherwise, the mail was successfully relayed at least once.
                relayed	       += 1
                log.info( f"Sent SMTP email via relay {mx} MAIL FROM: {from_addr} RCPT TO: {commas( to_addrs, final='and' )}" )
                break

    log.info( f"{relayed} of {len( relay_addrs )} relays succeeded" )

    return msg


class PostQueue:
    """A postfix-compatible post-queue filter.  See:
    https://codepoets.co.uk/2015/python-content_filter-for-postfix-rewriting-the-subject/

    Receives an email via stdin, and re-injects it into the mail system via /usr/sbin/sendmail,
    raising an exception if for any reason it is unable to do so (caller should trap and return
    an appropriate exit status compatible w/ postfix:

         0: success
        69: bounce
        75: tempfail

    Creates and sends a response email via SMTP to a relay (default is the local SMTP server at
    localhost:25)

    The msg, from_addr and to_addrs are retained in self.msg, etc.

    Postfix will pass all To:, Cc: and Bcc: recipients in to_addrs.

    The reinject command is executed/called, and passed from_addr and *to_addrs; False disables.
    """
    def __init__( self, reinject=None ):
        if reinject in (None, True):
            reinject		= [ '/usr/sbin/sendmail', '-G', '-i', '-f' ]
        if is_listlike( reinject ):
            self.reinject	= list( map( str, reinject ))   # [...]: command, eg. 'sendmail', ...
        elif not reinject:
            self.reinject	= lambda *args: None		# False, '': A NO-OP -- do not re-inject
        else:
            self.reinject	= reinject			# str, callable: a Shell command, or a function
        self.reset()

    def reset( self ):
        self.msg		= None				# The incoming original email.Message
        self.msg_peers		= (None,[])			# and its (mail_from, [rcpt_to, ...])
        self.rsp		= None				# The intended response email.Message
        self.rsp_peers		= (None,[])			# and its (mail_from, [rcpt_to, ...])

    def __call__( self, from_addr, *to_addrs ):
        """Respond to an incoming email.Message according to the filter.  Returns an appropriate Filter exit
        status code (see above).

        The base class simply re-injects the message, and does nothing else (does not even try to
        generate the self.rsp from self.response( msg ), which remains None.)

        """
        self.reset()
        self.msg		= self.message()
        self.msg_peers		= (from_addr, to_addrs)
        try:
            # Allow a command (list), a shell command or a function for reinject, logging
            # the results of the re-injection command/function.
            err			= None
            if is_listlike( self.reinject ):
                with Popen(
                        self.reinject + [ from_addr] + list( to_addrs ),
                        stdin	= PIPE,
                        stdout	= PIPE,
                        stderr	= PIPE
                ) as process:
                    out, err	= process.communicate( self.msg.as_bytes() )
                    out		= out.decode( 'UTF-8' )
                    err		= err.decode( 'UTF-8' )
            elif isinstance( self.reinject, str ):
                with Popen(
                        f"{self.reinject} {from_addr} {' '.join( to_addrs )}",
                        shell	= True,
                        stdin	= PIPE,
                        stdout	= PIPE,
                        stderr	= PIPE
                ) as process:
                    out, err	= process.communicate( self.msg.as_bytes() )
                    out		= out.decode( 'UTF-8' )
                    err		= err.decode( 'UTF-8' )
            else:
                out		= self.reinject( from_addr, *to_addrs )
            log.info( f"Results of reinjection: stdout: {out}, stderr: {err}" )
        except Exception as exc:
            # Something went wrong attempting re-injection
            log.warning( f"Re-injecting message From: {from_addr}, To: {commas( to_addrs )} via sendmail failed: {exc}" )
            return 75  # tempfail
        return 0

    def message( self ):
        """Obtain and return the email.Message we are to process.  By default, an filter receives its
        email.Message from stdin, unaltered.

        """
        return message_from_file( sys.stdin )

    def response( self, msg, new_id=None ):
        """Prepare the auto-responder message.  Copies the incoming email.Message, to avoid altering
        it. Optionally, give it a different Message-ID.

        """
        rsp			= message_from_string( msg.as_string() )
        if new_id:
            domain		= None
            # Prefer the sender field per RFC 2822:3.6.2.
            sender_m		= matchaddr( rsp['Sender'] if 'Sender' in rsp else rsp['From'] )
            if sender_m:
                domain		= sender_m[2]
            rsp['Message-ID']	= utils.make_msgid( domain=domain )
        return rsp


class AutoResponder( PostQueue ):
    """Filter DKIM-signed emails by Sender:/From: address, and auto-respond by sending a copy to all
    Cc/Bcc/Reply-To recipients.

    This is useful when incoming emails generated by an automated process have been DKIM-signed, and
    can therefore not have their To: field modified; there is no generally accepted method for delivering
    such emails to multiple parties.

    Once we are reasonably assured that this email is going to be delivered (eg. if it carries a
    financially costly payload, such as a Cryptocurrency wallet address, or the means to derive it),
    we may use this filter to (also) inform the intended recipient(s) (eg. in the Cc/Bcc/Reply-To)
    to proceed with the payment into the designated Cryptocurrency account.

    """
    def __init__( self, *args, address=None, server=None, port=None, **kwds ):
        address_m			= matchaddr( address or '' )
        assert address_m, \
            f"Must supply a valid email From: address to auto-respond to: {address}"
        self.address		= address
        self.mailbox,self.extension,self.domain	= address_m
        self.relay		= 'localhost' if server is None else server
        self.port		= 25 if port is None else port

        super().__init__( *args, **kwds )

        log.info( f"autoresponding to DKIM-signed emails From: {self.address}@{self.domain}" )

    def __call__( self, from_addr, *to_addrs ):
        """Decide if we should auto-respond, and do so.  Return the email.Message, and an appropriate
        Postfix exit code.

        """

        status			= super().__call__( from_addr, *to_addrs )
        if status:
            return status

        log.info( f"Filtered From: {self.msg['From']}, To: {self.msg['To']}"
                  f" originally     from {from_addr} to {len(to_addrs)} recipients: {commas( to_addrs, final='and' )}" )

        # Detect if this is a message we are intended to autorespond to; if not, do nothing.  First thing, ensure
        # that its Sender:/From: matches address (allows Glob wildcards).
        from_m			= matchaddr(
            self.msg['Sender'] or self.msg['From'],
            mailbox	= self.mailbox,
            extension	= self.extension,
            domain	= self.domain
        )
        if not from_m:
            log.warning( f"Message From: {self.msg['From']}, To: {self.msg['To']} expected Sender:/From: {self.address}; not auto-responding" )
            return 0
        if 'DKIM-Signature' not in self.msg:
            log.warning( f"Message From: {self.msg['From']}, To: {self.msg['To']} is not DKIM signed; not auto-responding" )
            return 0
        if not dkim.verify( self.msg.as_bytes() ):
            log.warning( f"Message From: {self.msg['From']}, To: {self.msg['To']} DKIM signature fails; not auto-responding " )
            return 0

        # This message is a target: Normalize, uniqueify and filter the addresses (discarding
        # invalids).  Avoid sending it to the designated To: address as this would likely set up a
        # mail loop.
        to_addr			= utils.parseaddr( self.msg['To'] )[1]  # May be empty, if no To: in header
        to_addrs_filt		= [
            fa
            for fa in uniq( filter( None, (
                utils.parseaddr( a )[1]
                for a in to_addrs
            )))
            if fa != to_addr
        ]

        # Get the outgoing message; also use the Reply-To: address, if supplied
        self.rsp		= self.response( self.msg )
        if 'Reply-To' in self.rsp:
            _,reply_to		= utils.parseaddr( self.rsp['Reply-To'] )
            if reply_to not in to_addrs_filt:
                to_addrs_filt.append( reply_to )
        self.rsp_peers		= (from_addr, tuple( to_addrs_filt ))
        log.info( f"Response From: {self.rsp['From']}, To: {self.rsp['To']}"
                  f" autoresponding from {from_addr} to {len(to_addrs_filt)} recipients: {commas( to_addrs_filt, final='and' )}"
                  + ( ""  if set( to_addrs ) == set( to_addrs_filt ) else f" (was {len(to_addrs)}: {commas( to_addrs, final='and' )})" ))
        if not to_addrs_filt:
            log.warning( f"Message From: {self.msg['From']}, To: {self.msg['To']} has no additional recipients; not auto-responding " )
            return 0

        # Now, send the same message to all the supplied Reply-To, and Cc/Bcc addresses (which were
        # already in to_addrs).  If it is DKIM signed, it will remain so, since we're not adjusting
        # the message -- just send w/ new RCPT TO: envelope addresses.  We'll use the same from_addr
        # address.
        try:
            send_message(
                self.rsp,
                from_addr	= from_addr,
                to_addrs	= to_addrs_filt,
                relay		= self.relay,
                port		= self.port,
            )
        except Exception as exc:
            log.warning( f"Message From: {self.rsp['From']}, To: {self.rsp['To']} autoresponder SMTP send failed: {exc}" )
            return 75  # tempfail

        return 0


#
# Provide a CLI to access the AutoResponder
#
@click.group()
@click.option('-v', '--verbose', count=True)
@click.option('-q', '--quiet', count=True)
@click.option( '--json/--no-json', default=True, help="Output JSON (the default)")
def cli( verbose, quiet, json ):
    cli.verbosity		= verbose - quiet
    log_cfg['level']		= log_level( cli.verbosity )
    logging.basicConfig( **log_cfg )
    if verbose or quiet:
        logging.getLogger().setLevel( log_cfg['level'] )
    cli.json			= json
cli.verbosity			= 0  # noqa: E305
cli.json			= False


@click.command()
@click.argument( 'address', nargs=1 )		# Only auto-respond to message with Sender:/From: this address
@click.argument( 'from_addr', nargs=1 )
@click.argument( 'to_addrs', nargs=-1 )
@click.option( '--server', default='localhost' )
@click.option( '--port', default=25 )
@click.option( '--reinject', type=str, default=None, help="A custom command to reinject email, eg. via sendmail" )
@click.option( '--no-reinject', 'reinject', flag_value='', help="Disable reinjection of filtered mail, eg. via sendmail" )
def autoresponder( address, from_addr, to_addrs, server, port, reinject ):
    """Run an auto-responder that replies to all incoming emails to the specified email address.

    Will be invoked with a from_addr and 1 or more to_addrs.

    - Must be DKIM signed, including the From: and To: addresses.
    - The RCPT TO: "envelope" address must match 'address':
      - We won't autorespond to copies of the email being delivered to other inboxes
    - The MAIL FROM: "envelope" address must match the From: address
      - We won't autorespond to copies forwarded from other email addresses


    Build/install Python3.10
     $ wget https://www.python.org/ftp/python/3.10.9/Python-3.10.9.tgz
     $ sudo apt-get -u  install wget build-essential libreadline-gplv2-dev libncursesw5-dev libssl-dev \
         libsqlite3-dev tk-dev libgdbm-dev libc6-dev libbz2-dev libffi-dev zlib1g-dev
     $ tar xzf Python-3.10.9.tgz
     $ cd Python-3.10.9
     $ ./configure --enable-optimizations --with-readline
     $ make
     $ make altinstall

    Install python-slip39 (no extras required for just slip39.communications)

      # rm -rf ~/.cache/pip
      # python3.10 -m pip install --upgrade https://github.com/pjkundert/python-slip39/archive/feature-invoice.zip#egg=slip39

    Configure Postfix system roughly as per: https://github.com/innovara/autoreply,
    to run our slip39.communications autoresponder

    - Create 'autoreply' user, and /opt/autoreply (not presently used)
      - Only necessary if your filter uses filesystem; otherwise, use 'nobody'

    - Configure /etc/postfix/master.cf to run the autoreply filter

        # autoresponder pipe
        autoreply unix  -       n       n       -       -       pipe
         flags= user=nobody null_sender=
         argv=/usr/local/bin/python3.10 -m slip39.communications autoresponder licensing@dominionrnd.com ${sender} ${recipient}

    - Create /etc/postfix/autoreply w/ lines like (and postmap /etc/postfix/autoreply):

        licensing@dominionrnd.com	FILTER	autoreply:dummy

    - Configure /etc/postfix/main.cf to process the autoreply filter

        # Trigger the autoreply filter for specified recipient addresses
        smtpd_recipient_restrictions =
            check_recipient_access hash:/etc/postfix/autoreply

    """
    ar				= AutoResponder(
        address		= address,
        server		= server,
        port		= port,
        reinject	= reinject,
    )
    status			= ar( from_addr, *to_addrs )
    if cli.json and cli.verbosity:
        click.echo( json.dumps( dict(
            dict(
                dict(
                    Original	= ar.msg.as_string() if ar.msg else None,
                ) if cli.verbosity > 2 else dict(),
                Response	= ar.rsp.as_string() if ar.rsp else None,
            ) if cli.verbosity > 1 else dict(),
            status		= status,
            MAIL_FROM		= ar.rsp_peers[0],
            RCPT_TOs		= ar.rsp_peers[1],
        ), indent=4 ))
    elif cli.verbosity:
        click.echo( f"status:    {status}" )
        click.echo( f"MAIL FROM: {ar.rsp_peers[0]}" )
        click.echo( f"RCPT TOs:  {commas( ar.rsp_peers[1] )}" )
        if cli.verbosity > 2:
            click.echo( f"Original:\n{ar.msg.as_string() if ar.msg else None}" )
        if cli.verbosity > 1:
            click.echo( f"Response:\n{ar.rsp.as_string() if ar.rsp else None}" )
    sys.exit( status )


cli.add_command( autoresponder )


if __name__ == "__main__":
    cli()
